"""
Multicomponent kernels (simple) restrict all signal variance and length scale of hyperparameters
to a single value. The kernels in this module allow you to have different sets of hyperparameters
and cutoffs for different interactions, and have flexible groupings of elements. It also allows
you to do partial hyper-parameter training, keeping some components fixed.

To use this set of kernels, we need a hyps_mask dictionary for GaussianProcess, MappedGaussianProcess,
and AtomicEnvironment (if you also set up different cutoffs).  A simple example is shown below.

Examples:

    >>> from flare.util.parameter_helper import ParameterHelper
    >>> from flare.gp import GaussianProcess

    >>> pm = ParameterHelper(species=['O', 'C', 'H'],
    ...                      kernels={'twobody':[['*', '*'], ['O','O']],
    ...                               'threebody':[['*', '*', '*'], ['O','O', 'O']]},
    ...                      parameters={'twobody0':[1, 0.5, 1], 'twobody1':[2, 0.2, 2],
    ...                            'triplet0':[1, 0.5], 'triplet1':[2, 0.2],
    ...                            'cutoff_twobody':2, 'cutoff_threebody':1, 'noise': 0.05},
    ...                      constraints={'twobody0':[False, True]})
    >>> hyps_mask = pm1.as_dict()
    >>> hyps = hyps_mask.pop('hyps')
    >>> cutoffs = hyps_mask.pop('cutoffs')
    >>> hyp_labels = hyps_mask.pop('hyp_labels')
    >>> kernels = hyps_mask['kernels']
    >>> gp_model = GaussianProcess(kernels=kernels,
    ...                            hyps=hyps, cutoffs=cutoffs,
    ...                            hyp_labels=hyp_labels,
    ...                            parallel=True, per_atom_par=False,
    ...                            n_cpus=n_cpus,
    ...                            multihyps=True, hyps_mask=hm)


In the example above, Parameters class generates the arrays needed
for these kernels and store all the grouping and mapping information in the
hyps_mask dictionary.  It stores following keys and values:

* spec_mask: 118-long integer array descirbing which elements belong to
             like groups for determining which bond hyperparameters to use. For
             instance, [0,0,1,1,0 ...] assigns H to group 0, He and Li to group 1,
             and Be to group 0 (the 0th register is ignored).
* nspec: Integer, number of different species groups (equal to number of
         unique values in spec_mask).
* nbond: Integer, number of different hyperparameter sets to associate with
         different 2-body pairings of atoms in groups defined in spec_mask.
* bond_mask: Array of length nspec^2, which describes the hyperparameter sets to
             associate with different pairings of species types. For example, if there
             are atoms of type 0 and 1, then bond_mask defines which hyperparameters
             to use for parings [0-0, 0-1, 1-0, 1-1]: if we wanted hyperparameter set 0 for
             0-0 parings and set 1 for 0-1 and 1-1 pairings, then we would make
             bond_mask [0, 1, 1, 1].
* ntriplet: Integer, number of different hyperparameter sets to associate
            with different 3-body pariings of atoms in groups defined in spec_mask.
* triplet_mask: Similar to bond mask: Triplet pairings of type 0 and 1 atoms
                would go {0-0-0, 0-0-1, 0-1-0, 0-1-1, 1-0-0, 1-0-1, 1-1-0, 1-1-1},
                and if we wanted hyp. set 0 for triplets with only atoms of type 0
                and hyp. set 1 for all the rest, then the triplet_mask array would
                read [0,1,1,1,1,1,1,1]. The user should make sure that the mask has
                a permutational symmetry.
* cutoff_2b: Array of length nbond, which stores the cutoff used for different
             types of bonds defined in bond_mask
* ncut3b:    Integer, number of different cutoffs sets to associate
             with different 3-body pariings of atoms in groups defined in spec_mask.
* cut3b_mask: Array of length nspec^2, which describes the cutoff to
             associate with different bond types in triplets. For example, in a triplet
             (C, O, H) , there are three cutoffs. Cutoffs for CH bond, CO bond and OH bond.
             If C and O are associate with atom group 1 in spec_mask and H are associate with
             group 0 in spec_mask, the cut3b_mask[1*nspec+0] determines the C/O-H bond cutoff,
             and cut3b_mask[1*nspec+1] determines the C-O bond cutoff. If we want the
             former one to use the 1st cutoff in cutoff_3b and the later to use the 2nd cutoff
             in cutoff_3b, the cut3b_mask should be [0, 0, 0, 1]
* cutoff_3b: Array of length ncut3b, which stores the cutoff used for different
             types of bonds in triplets.
* nmb :      Integer, number of different cutoffs set to associate with different coordination
             numbers
* mb_mask:   similar to bond_mask and cut3b_mask.
* cutoff_mb: Array of length nmb, stores the cutoff used for different many body terms

For selective optimization. one can define 'map', 'train_noise' and 'original'
to identify which element to be optimized. All three have to be defined.
train_noise = Bool (True/False), whether the noise parameter can be optimized
original: np.array. Full set of initial values for hyperparmeters
map: np.array, array to map the hyper parameter back to the full set.
map[i]=j means the i-th element in hyps should be the j-th element in
hyps_mask['original']

For example, the full set of hyper parmeters
may include [ls21, ls22, sig21, sig22, ls3
sg3, noise] but suppose you wanted only the set 21 optimized.
The full set of hyperparameters is defined in 'original'; include all those
you want to leave static, and set initial guesses for those you want to vary.
Have the 'map' list contain the indices of the hyperparameters in 'original'
that correspond to the hyperparameters you want to vary.
Have a hyps list which contain those which you want to vary. Below,
ls21, ls22 etc... represent floating-point variables which correspond
to the initial guesses / static values.
You would then pass in:

hyps = [ls21, sig21]
hyps_mask = { ..., 'train_noise': False, 'map':[0, 2],
                   'original': [ls21, ls22, sig21, sig22, ls3, sg3, noise]}
the hyps argument should only contain the values that need to be optimized.
If you want noise to be trained as well include noise as the
final hyperparameter value in hyps.

"""

from math import exp
import numpy as np

from numba import njit

import flare.kernels.cutoffs as cf

from flare.kernels.kernels import force_helper, grad_constants, grad_helper, \
    force_energy_helper, three_body_en_helper, three_body_helper_1, \
    three_body_helper_2, three_body_grad_helper_1, three_body_grad_helper_2
from flare.kernels.mc_3b_sepcut import three_body_mc_sepcut_jit, \
    three_body_mc_grad_sepcut_jit, three_body_mc_force_en_sepcut_jit, \
    three_body_mc_en_sepcut_jit
from flare.kernels.mc_mb_sepcut import \
    many_body_mc_sepcut_jit, many_body_mc_grad_sepcut_jit, \
    many_body_mc_force_en_sepcut_jit, many_body_mc_en_sepcut_jit

# -----------------------------------------------------------------------------
#                        two plus three plus many body kernels
# -----------------------------------------------------------------------------


def two_three_many_body_mc(env1, env2, d1, d2, cutoff_2b, cutoff_3b, cutoff_mb,
                           nspec, spec_mask,
                           nbond, bond_mask, ntriplet, triplet_mask,
                           ncut3b, cut3b_mask,
                           nmb, mb_mask,
                           sig2, ls2, sig3, ls3, sigm, lsm,
                           cutoff_func=cf.quadratic_cutoff):
    """2+3+manybody multi-element kernel between two force components.

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b (float, np.ndarray): cutoff(s) for two-body interaction
        cutoff_3b (float, np.ndarray): cutoff(s) for three-body interaction
        cutoff_mb (float, np.ndarray): cutoff(s) for coordination-based manybody interaction
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond (int): number of different hyperparameter sets to associate with two-body pairings
        bond_mask (np.ndarray): nspec^2 long integer array
        ntriplet (int): number of different hyperparameter sets to associate with 3-body pairings
        triplet_mask (np.ndarray): nspec^3 long integer array
        ncut3b (int): number of different 3-body cutoff sets to associate with 3-body pairings
        cut3b_mask (np.ndarray): nspec^2 long integer array
        nmb (int): number of different hyperparameter sets to associate with manybody pairings
        mb_mask (np.ndarray): nspec^2 long integer array
        sig2 (np.ndarray): signal variances associates with two-body term
        ls2 (np.ndarray): length scales associates with two-body term
        sig3 (np.ndarray): signal variances associates with three-body term
        ls3 (np.ndarray): length scales associates with three-body term
        sigm (np.ndarray): signal variances associates with manybody term
        lsm (np.ndarray): length scales associates with manybody term
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        float: Value of the 2+3+many-body kernel.
    """

    two_term = two_body_mc_jit(env1.bond_array_2, env1.ctype, env1.etypes,
                               env2.bond_array_2, env2.ctype, env2.etypes,
                               d1, d2, sig2, ls2, cutoff_2b, cutoff_func,
                               nspec, spec_mask, bond_mask)

    if (ncut3b == 0):
        tbmcj = three_body_mc_jit
    else:
        tbmcj = three_body_mc_sepcut_jit

    three_term = \
        tbmcj(env1.bond_array_3, env1.ctype, env1.etypes,
              env2.bond_array_3, env2.ctype, env2.etypes,
              env1.cross_bond_inds, env2.cross_bond_inds,
              env1.cross_bond_dists, env2.cross_bond_dists,
              env1.triplet_counts, env2.triplet_counts,
              d1, d2, sig3, ls3, cutoff_3b, cutoff_func,
              nspec, spec_mask, triplet_mask, cut3b_mask)

    mbmcj = many_body_mc_sepcut_jit
    many_term = mbmcj(env1.q_array, env2.q_array,
                      env1.q_neigh_array, env2.q_neigh_array,
                      env1.q_neigh_grads, env2.q_neigh_grads,
                      env1.ctype, env2.ctype,
                      env1.etypes_mb, env2.etypes_mb,
                      env1.unique_species, env2.unique_species,
                      d1, d2, sigm, lsm,
                      nspec, spec_mask, mb_mask)

    return two_term + three_term + many_term


def two_three_many_body_mc_grad(env1, env2, d1, d2, cutoff_2b, cutoff_3b, cutoff_mb,
                                nspec, spec_mask,
                                nbond, bond_mask, ntriplet, triplet_mask,
                                ncut3b, cut3b_mask,
                                nmb, mb_mask,
                                sig2, ls2, sig3, ls3, sigm, lsm,
                                cutoff_func=cf.quadratic_cutoff):
    """2+3+manybody multi-element kernel between two force components and its
    gradient with respect to the hyperparameters.

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b (float, np.ndarray): cutoff(s) for two-body interaction
        cutoff_3b (float, np.ndarray): cutoff(s) for three-body interaction
        cutoff_mb (float, np.ndarray): cutoff(s) for coordination-based manybody interaction
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond (int): number of different hyperparameter sets to associate with two-body pairings
        bond_mask (np.ndarray): nspec^2 long integer array
        ntriplet (int): number of different hyperparameter sets to associate with 3-body pairings
        triplet_mask (np.ndarray): nspec^3 long integer array
        ncut3b (int): number of different 3-body cutoff sets to associate with 3-body pairings
        cut3b_mask (np.ndarray): nspec^2 long integer array
        nmb (int): number of different hyperparameter sets to associate with manybody pairings
        mb_mask (np.ndarray): nspec^2 long integer array
        sig2 (np.ndarray): signal variances associates with two-body term
        ls2 (np.ndarray): length scales associates with two-body term
        sig3 (np.ndarray): signal variances associates with three-body term
        ls3 (np.ndarray): length scales associates with three-body term
        sigm (np.ndarray): signal variances associates with manybody term
        lsm (np.ndarray): length scales associates with manybody term
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        (float, np.ndarray):
            Value of the 2+3+manybody kernel and its gradient
            with respect to the hyperparameters.
    """

    kern2, grad2 = \
        two_body_mc_grad_jit(env1.bond_array_2, env1.ctype, env1.etypes,
                             env2.bond_array_2, env2.ctype, env2.etypes,
                             d1, d2, sig2, ls2, cutoff_2b, cutoff_func,
                             nspec, spec_mask,
                             nbond, bond_mask)

    if (ncut3b == 0):
        tbmcj = three_body_mc_grad_jit
    else:
        tbmcj = three_body_mc_grad_sepcut_jit

    kern3, grad3 = \
        tbmcj(env1.bond_array_3, env1.ctype, env1.etypes,
              env2.bond_array_3, env2.ctype, env2.etypes,
              env1.cross_bond_inds, env2.cross_bond_inds,
              env1.cross_bond_dists, env2.cross_bond_dists,
              env1.triplet_counts, env2.triplet_counts,
              d1, d2, sig3, ls3, cutoff_3b,
              cutoff_func,
              nspec, spec_mask,
              ntriplet, triplet_mask, cut3b_mask)

    mbmcj = many_body_mc_grad_sepcut_jit
    kern_many, gradm = mbmcj(env1.q_array, env2.q_array,
                             env1.q_neigh_array, env2.q_neigh_array,
                             env1.q_neigh_grads, env2.q_neigh_grads,
                             env1.ctype, env2.ctype,
                             env1.etypes_mb, env2.etypes_mb,
                             env1.unique_species, env2.unique_species,
                             d1, d2, sigm, lsm,
                             nspec, spec_mask, nmb, mb_mask)

    return kern2 + kern3 + kern_many, np.hstack([grad2, grad3, gradm])


def two_three_many_mc_force_en(env1, env2, d1, cutoff_2b, cutoff_3b, cutoff_mb,
                               nspec, spec_mask,
                               nbond, bond_mask, ntriplet, triplet_mask,
                               ncut3b, cut3b_mask,
                               nmb, mb_mask,
                               sig2, ls2, sig3, ls3, sigm, lsm,
                               cutoff_func=cf.quadratic_cutoff):
    """2+3+manybody multi-element kernel between a force component and a local
    energy.

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b (float, np.ndarray): cutoff(s) for two-body interaction
        cutoff_3b (float, np.ndarray): cutoff(s) for three-body interaction
        cutoff_mb (float, np.ndarray): cutoff(s) for coordination-based manybody interaction
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond (int): number of different hyperparameter sets to associate with two-body pairings
        bond_mask (np.ndarray): nspec^2 long integer array
        ntriplet (int): number of different hyperparameter sets to associate with 3-body pairings
        triplet_mask (np.ndarray): nspec^3 long integer array
        ncut3b (int): number of different 3-body cutoff sets to associate with 3-body pairings
        cut3b_mask (np.ndarray): nspec^2 long integer array
        nmb (int): number of different hyperparameter sets to associate with manybody pairings
        mb_mask (np.ndarray): nspec^2 long integer array
        sig2 (np.ndarray): signal variances associates with two-body term
        ls2 (np.ndarray): length scales associates with two-body term
        sig3 (np.ndarray): signal variances associates with three-body term
        ls3 (np.ndarray): length scales associates with three-body term
        sigm (np.ndarray): signal variances associates with manybody term
        lsm (np.ndarray): length scales associates with manybody term
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        float: Value of the 2+3+many-body force/energy kernel.
    """

    two_term = \
        two_body_mc_force_en_jit(env1.bond_array_2, env1.ctype, env1.etypes,
                                 env2.bond_array_2, env2.ctype, env2.etypes,
                                 d1, sig2, ls2, cutoff_2b, cutoff_func,
                                 nspec, spec_mask,
                                 bond_mask) / 2

    if (ncut3b == 0):
        tbmcj = three_body_mc_force_en_jit
    else:
        tbmcj = three_body_mc_force_en_sepcut_jit

    three_term = \
        tbmcj(env1.bond_array_3, env1.ctype, env1.etypes,
              env2.bond_array_3, env2.ctype, env2.etypes,
              env1.cross_bond_inds, env2.cross_bond_inds,
              env1.cross_bond_dists,
              env2.cross_bond_dists,
              env1.triplet_counts, env2.triplet_counts,
              d1, sig3, ls3, cutoff_3b, cutoff_func,
              nspec, spec_mask, triplet_mask,
              cut3b_mask) / 3

    mbmcj = many_body_mc_force_en_sepcut_jit
    many_term = mbmcj(env1.q_array, env2.q_array,
                      env1.q_neigh_array, env1.q_neigh_grads,
                      env1.ctype, env2.ctype, env1.etypes_mb,
                      env1.unique_species, env2.unique_species,
                      d1, sigm, lsm,
                      nspec, spec_mask, mb_mask)

    return two_term + three_term + many_term


def two_three_many_mc_en(env1, env2, cutoff_2b, cutoff_3b, cutoff_mb,
                         nspec, spec_mask,
                         nbond, bond_mask, ntriplet, triplet_mask,
                         ncut3b, cut3b_mask,
                         nmb, mb_mask,
                         sig2, ls2, sig3, ls3, sigm, lsm,
                         cutoff_func=cf.quadratic_cutoff):
    """2+3+many-body multi-element kernel between two local energies.

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b (float, np.ndarray): cutoff(s) for two-body interaction
        cutoff_3b (float, np.ndarray): cutoff(s) for three-body interaction
        cutoff_mb (float, np.ndarray): cutoff(s) for coordination-based manybody interaction
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond (int): number of different hyperparameter sets to associate with two-body pairings
        bond_mask (np.ndarray): nspec^2 long integer array
        ntriplet (int): number of different hyperparameter sets to associate with 3-body pairings
        triplet_mask (np.ndarray): nspec^3 long integer array
        ncut3b (int): number of different 3-body cutoff sets to associate with 3-body pairings
        cut3b_mask (np.ndarray): nspec^2 long integer array
        nmb (int): number of different hyperparameter sets to associate with manybody pairings
        mb_mask (np.ndarray): nspec^2 long integer array
        sig2 (np.ndarray): signal variances associates with two-body term
        ls2 (np.ndarray): length scales associates with two-body term
        sig3 (np.ndarray): signal variances associates with three-body term
        ls3 (np.ndarray): length scales associates with three-body term
        sigm (np.ndarray): signal variances associates with manybody term
        lsm (np.ndarray): length scales associates with manybody term
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        float: Value of the 2+3+many-body energy/energy kernel.
    """

    two_term = two_body_mc_en_jit(env1.bond_array_2, env1.ctype, env1.etypes,
                                  env2.bond_array_2, env2.ctype, env2.etypes,
                                  sig2, ls2, cutoff_2b, cutoff_func,
                                  nspec,
                                  spec_mask,
                                  bond_mask)/4

    if (ncut3b == 0):
        tbmcj = three_body_mc_en_jit
    else:
        tbmcj = three_body_mc_en_sepcut_jit

    three_term = \
        tbmcj(env1.bond_array_3, env1.ctype, env1.etypes,
              env2.bond_array_3, env2.ctype, env2.etypes,
              env1.cross_bond_inds, env2.cross_bond_inds,
              env1.cross_bond_dists, env2.cross_bond_dists,
              env1.triplet_counts, env2.triplet_counts,
              sig3, ls3, cutoff_3b, cutoff_func,
              nspec, spec_mask,
              triplet_mask, cut3b_mask)/9.

    mbmcj = many_body_mc_en_sepcut_jit
    many_term = mbmcj(env1.q_array, env2.q_array,
                      env1.ctype, env2.ctype,
                      env1.unique_species, env2.unique_species,
                      sigm, lsm,
                      nspec, spec_mask, mb_mask)

    return two_term + three_term + many_term


# -----------------------------------------------------------------------------
#                        two plus three body kernels
# -----------------------------------------------------------------------------


def two_plus_three_body_mc(env1, env2, d1, d2, cutoff_2b, cutoff_3b, cutoff_mb,
                           nspec, spec_mask,
                           nbond, bond_mask, ntriplet, triplet_mask,
                           ncut3b, cut3b_mask,
                           nmb, mb_mask,
                           sig2, ls2, sig3, ls3, sigm, lsm,
                           cutoff_func=cf.quadratic_cutoff):
    """2+3-body multi-element kernel between two force components.

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b (float, np.ndarray): cutoff(s) for two-body interaction
        cutoff_3b (float, np.ndarray): cutoff(s) for three-body interaction
        cutoff_mb (float, np.ndarray): cutoff(s) for coordination-based manybody interaction
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond (int): number of different hyperparameter sets to associate with two-body pairings
        bond_mask (np.ndarray): nspec^2 long integer array
        ntriplet (int): number of different hyperparameter sets to associate with 3-body pairings
        triplet_mask (np.ndarray): nspec^3 long integer array
        ncut3b (int): number of different 3-body cutoff sets to associate with 3-body pairings
        cut3b_mask (np.ndarray): nspec^2 long integer array
        sig2 (np.ndarray): signal variances associates with two-body term
        ls2 (np.ndarray): length scales associates with two-body term
        sig3 (np.ndarray): signal variances associates with three-body term
        ls3 (np.ndarray): length scales associates with three-body term
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        float: Value of the 2+3-body force/force kernel.
    """

    two_term = two_body_mc_jit(env1.bond_array_2, env1.ctype, env1.etypes,
                               env2.bond_array_2, env2.ctype, env2.etypes,
                               d1, d2, sig2, ls2, cutoff_2b, cutoff_func,
                               nspec, spec_mask, bond_mask)

    if (ncut3b <= 1):
        tbmcj = three_body_mc_jit
    else:
        tbmcj = three_body_mc_sepcut_jit

    three_term = \
        tbmcj(env1.bond_array_3, env1.ctype, env1.etypes,
              env2.bond_array_3, env2.ctype, env2.etypes,
              env1.cross_bond_inds, env2.cross_bond_inds,
              env1.cross_bond_dists, env2.cross_bond_dists,
              env1.triplet_counts, env2.triplet_counts,
              d1, d2, sig3, ls3, cutoff_3b, cutoff_func,
              nspec, spec_mask, triplet_mask, cut3b_mask)

    return two_term + three_term


def two_plus_three_body_mc_grad(env1, env2, d1, d2, cutoff_2b, cutoff_3b, cutoff_mb,
                                nspec, spec_mask,
                                nbond, bond_mask, ntriplet, triplet_mask,
                                ncut3b, cut3b_mask,
                                nmb, mb_mask,
                                sig2, ls2, sig3, ls3, sigm, lsm,
                                cutoff_func=cf.quadratic_cutoff):
    """2+3-body multi-element kernel between two force components and its
    gradient with respect to the hyperparameters.

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b (float, np.ndarray): cutoff(s) for two-body interaction
        cutoff_3b (float, np.ndarray): cutoff(s) for three-body interaction
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond (int): number of different hyperparameter sets to associate with two-body pairings
        bond_mask (np.ndarray): nspec^2 long integer array
        ntriplet (int): number of different hyperparameter sets to associate with 3-body pairings
        triplet_mask (np.ndarray): nspec^3 long integer array
        ncut3b (int): number of different 3-body cutoff sets to associate with 3-body pairings
        cut3b_mask (np.ndarray): nspec^2 long integer array
        sig2 (np.ndarray): signal variances associates with two-body term
        ls2 (np.ndarray): length scales associates with two-body term
        sig3 (np.ndarray): signal variances associates with three-body term
        ls3 (np.ndarray): length scales associates with three-body term
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        (float, np.ndarray):
            Value of the 2+3-body kernel and its gradient
            with respect to the hyperparameters.
    """

    kern2, grad2 = \
        two_body_mc_grad_jit(env1.bond_array_2, env1.ctype, env1.etypes,
                             env2.bond_array_2, env2.ctype, env2.etypes,
                             d1, d2, sig2, ls2, cutoff_2b, cutoff_func,
                             nspec, spec_mask,
                             nbond, bond_mask)

    if (ncut3b == 0):
        tbmcj = three_body_mc_grad_jit
    else:
        tbmcj = three_body_mc_grad_sepcut_jit

    kern3, grad3 = \
        tbmcj(env1.bond_array_3, env1.ctype, env1.etypes,
              env2.bond_array_3, env2.ctype, env2.etypes,
              env1.cross_bond_inds, env2.cross_bond_inds,
              env1.cross_bond_dists, env2.cross_bond_dists,
              env1.triplet_counts, env2.triplet_counts,
              d1, d2, sig3, ls3, cutoff_3b,
              cutoff_func,
              nspec, spec_mask,
              ntriplet, triplet_mask, cut3b_mask)

    g = np.hstack([grad2, grad3])

    return kern2 + kern3, g


def two_plus_three_mc_force_en(env1, env2, d1, cutoff_2b, cutoff_3b, cutoff_mb,
                               nspec, spec_mask, nbond, bond_mask,
                               ntriplet, triplet_mask, ncut3b, cut3b_mask,
                               nmb, mb_mask,
                               sig2, ls2, sig3, ls3, sigm, lsm,
                               cutoff_func=cf.quadratic_cutoff):
    """2+3-body multi-element kernel between force and local energy

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b (float, np.ndarray): cutoff(s) for two-body interaction
        cutoff_3b (float, np.ndarray): cutoff(s) for three-body interaction
        cutoff_mb (float, np.ndarray): cutoff(s) for coordination-based manybody interaction
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond (int): number of different hyperparameter sets to associate with two-body pairings
        bond_mask (np.ndarray): nspec^2 long integer array
        ntriplet (int): number of different hyperparameter sets to associate with 3-body pairings
        triplet_mask (np.ndarray): nspec^3 long integer array
        ncut3b (int): number of different 3-body cutoff sets to associate with 3-body pairings
        cut3b_mask (np.ndarray): nspec^2 long integer array
        sig2 (np.ndarray): signal variances associates with two-body term
        ls2 (np.ndarray): length scales associates with two-body term
        sig3 (np.ndarray): signal variances associates with three-body term
        ls3 (np.ndarray): length scales associates with three-body term
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        float: Value of the 2+3-body force/energy kernel.
    """

    two_term = \
        two_body_mc_force_en_jit(env1.bond_array_2, env1.ctype, env1.etypes,
                                 env2.bond_array_2, env2.ctype, env2.etypes,
                                 d1, sig2, ls2, cutoff_2b, cutoff_func,
                                 nspec, spec_mask,
                                 bond_mask) / 2

    if (ncut3b == 0):
        tbmcj = three_body_mc_force_en_jit
    else:
        tbmcj = three_body_mc_force_en_sepcut_jit

    three_term = \
        tbmcj(env1.bond_array_3, env1.ctype, env1.etypes,
              env2.bond_array_3, env2.ctype, env2.etypes,
              env1.cross_bond_inds, env2.cross_bond_inds,
              env1.cross_bond_dists,
              env2.cross_bond_dists,
              env1.triplet_counts, env2.triplet_counts,
              d1, sig3, ls3, cutoff_3b, cutoff_func,
              nspec, spec_mask,
              triplet_mask, cut3b_mask) / 3

    return two_term + three_term


def two_plus_three_mc_en(env1, env2, cutoff_2b, cutoff_3b, cutoff_mb,
                         nspec, spec_mask, nbond, bond_mask,
                         ntriplet, triplet_mask, ncut3b, cut3b_mask,
                         nmb, mb_mask,
                         sig2, ls2, sig3, ls3, sigm, lsm,
                         cutoff_func=cf.quadratic_cutoff):
    """2+3-body multi-element kernel between two local energies

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b (float, np.ndarray): cutoff(s) for two-body interaction
        cutoff_3b (float, np.ndarray): cutoff(s) for three-body interaction
        cutoff_mb (float, np.ndarray): cutoff(s) for coordination-based manybody interaction
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond (int): number of different hyperparameter sets to associate with two-body pairings
        bond_mask (np.ndarray): nspec^2 long integer array
        ntriplet (int): number of different hyperparameter sets to associate with 3-body pairings
        triplet_mask (np.ndarray): nspec^3 long integer array
        ncut3b (int): number of different 3-body cutoff sets to associate with 3-body pairings
        cut3b_mask (np.ndarray): nspec^2 long integer array
        sig2 (np.ndarray): signal variances associates with two-body term
        ls2 (np.ndarray): length scales associates with two-body term
        sig3 (np.ndarray): signal variances associates with three-body term
        ls3 (np.ndarray): length scales associates with three-body term
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        float: Value of the 2+3-body energy/energy kernel.
    """

    two_term = two_body_mc_en_jit(env1.bond_array_2, env1.ctype, env1.etypes,
                                  env2.bond_array_2, env2.ctype, env2.etypes,
                                  sig2, ls2, cutoff_2b, cutoff_func,
                                  nspec,
                                  spec_mask,
                                  bond_mask) / 4

    if (ncut3b == 0):
        tbmcj = three_body_mc_en_jit
    else:
        tbmcj = three_body_mc_en_sepcut_jit

    three_term = \
        tbmcj(env1.bond_array_3, env1.ctype, env1.etypes,
              env2.bond_array_3, env2.ctype, env2.etypes,
              env1.cross_bond_inds, env2.cross_bond_inds,
              env1.cross_bond_dists, env2.cross_bond_dists,
              env1.triplet_counts, env2.triplet_counts,
              sig3, ls3, cutoff_3b, cutoff_func,
              nspec, spec_mask,
              triplet_mask, cut3b_mask)/9

    return two_term + three_term


# -----------------------------------------------------------------------------
#                      three body multicomponent kernel
# -----------------------------------------------------------------------------


def three_body_mc(env1, env2, d1, d2, cutoff_2b, cutoff_3b, cutoff_mb,
                  nspec, spec_mask, nbond, bond_mask,
                  ntriplet, triplet_mask, ncut3b, cut3b_mask,
                  nmb, mb_mask,
                  sig2, ls2, sig3, ls3, sigm, lsm,
                  cutoff_func=cf.quadratic_cutoff):
    """3-body multi-element kernel between two force components.

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b: dummy
        cutoff_3b (float, np.ndarray): cutoff(s) for three-body interaction
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond: dummy
        bond_mask: dummy
        ntriplet (int): number of different hyperparameter sets to associate with 3-body pairings
        triplet_mask (np.ndarray): nspec^3 long integer array
        ncut3b (int): number of different 3-body cutoff sets to associate with 3-body pairings
        cut3b_mask (np.ndarray): nspec^2 long integer array
        sig2: dummy
        ls2: dummy
        sig3 (np.ndarray): signal variances associates with three-body term
        ls3 (np.ndarray): length scales associates with three-body term
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        float: Value of the 2+3-body force/force kernel.
    """

    if (ncut3b == 0):
        tbmcj = three_body_mc_jit
    else:
        tbmcj = three_body_mc_sepcut_jit

    return tbmcj(env1.bond_array_3, env1.ctype, env1.etypes,
                 env2.bond_array_3, env2.ctype, env2.etypes,
                 env1.cross_bond_inds, env2.cross_bond_inds,
                 env1.cross_bond_dists, env2.cross_bond_dists,
                 env1.triplet_counts, env2.triplet_counts,
                 d1, d2, sig3, ls3, cutoff_3b, cutoff_func,
                 nspec, spec_mask,
                 triplet_mask, cut3b_mask)


def three_body_mc_grad(env1, env2, d1, d2, cutoff_2b, cutoff_3b, cutoff_mb,
                       nspec, spec_mask, nbond, bond_mask,
                       ntriplet, triplet_mask, ncut3b, cut3b_mask,
                       nmb, mb_mask,
                       sig2, ls2, sig3, ls3, sigm, lsm,
                       cutoff_func=cf.quadratic_cutoff):
    """3-body multi-element kernel between two force components and its
    gradient with respect to the hyperparameters.

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b: dummy
        cutoff_3b (float, np.ndarray): cutoff(s) for three-body interaction
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond: dummy
        bond_mask: dummy
        ntriplet (int): number of different hyperparameter sets to associate with 3-body pairings
        triplet_mask (np.ndarray): nspec^3 long integer array
        ncut3b (int): number of different 3-body cutoff sets to associate with 3-body pairings
        cut3b_mask (np.ndarray): nspec^2 long integer array
        sig2: dummy
        ls2: dummy
        sig3 (np.ndarray): signal variances associates with three-body term
        ls3 (np.ndarray): length scales associates with three-body term
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        (float, np.ndarray):
            Value of the 2+3+manybody kernel and its gradient
            with respect to the hyperparameters.
    """

    if (ncut3b == 0):
        tbmcj = three_body_mc_grad_jit
    else:
        tbmcj = three_body_mc_grad_sepcut_jit

    return tbmcj(
        env1.bond_array_3, env1.ctype, env1.etypes,
        env2.bond_array_3, env2.ctype, env2.etypes,
        env1.cross_bond_inds, env2.cross_bond_inds,
        env1.cross_bond_dists, env2.cross_bond_dists,
        env1.triplet_counts, env2.triplet_counts,
        d1, d2, sig3, ls3, cutoff_3b, cutoff_func,
        nspec, spec_mask, ntriplet, triplet_mask, cut3b_mask)


def three_body_mc_force_en(env1, env2, d1, cutoff_2b, cutoff_3b, cutoff_mb,
                           nspec, spec_mask, nbond, bond_mask, ntriplet, triplet_mask,
                           ncut3b, cut3b_mask, nmb, mb_mask,
                           sig2, ls2, sig3, ls3, sigm, lsm,
                           cutoff_func=cf.quadratic_cutoff):
    """3-body multi-element kernel between a force component and local energies

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b: dummy
        cutoff_3b (float, np.ndarray): cutoff(s) for three-body interaction
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond: dummy
        bond_mask: dummy
        ntriplet (int): number of different hyperparameter sets to associate with 3-body pairings
        triplet_mask (np.ndarray): nspec^3 long integer array
        ncut3b (int): number of different 3-body cutoff sets to associate with 3-body pairings
        cut3b_mask (np.ndarray): nspec^2 long integer array
        sig2: dummy
        ls2: dummy
        sig3 (np.ndarray): signal variances associates with three-body term
        ls3 (np.ndarray): length scales associates with three-body term
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        float: Value of the 2+3-body force/energy kernel.
    """

    if (ncut3b == 0):
        tbmcj = three_body_mc_force_en_jit
    else:
        tbmcj = three_body_mc_force_en_sepcut_jit

    return tbmcj(env1.bond_array_3, env1.ctype,
                 env1.etypes,
                 env2.bond_array_3, env2.ctype,
                 env2.etypes,
                 env1.cross_bond_inds,
                 env2.cross_bond_inds,
                 env1.cross_bond_dists,
                 env2.cross_bond_dists,
                 env1.triplet_counts,
                 env2.triplet_counts,
                 d1, sig3, ls3, cutoff_3b,
                 cutoff_func,
                 nspec,
                 spec_mask,
                 triplet_mask, cut3b_mask) / 3


def three_body_mc_en(env1, env2, cutoff_2b, cutoff_3b,  cutoff_mb,  nspec, spec_mask,
                     nbond, bond_mask, ntriplet, triplet_mask,
                     ncut3b, cut3b_mask, nmb, mb_mask,
                     sig2, ls2, sig3, ls3, sigm, lsm,
                     cutoff_func=cf.quadratic_cutoff):
    """3-body multi-element kernel between two local energies

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b: dummy
        cutoff_3b (float, np.ndarray): cutoff(s) for three-body interaction
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond: dummy
        bond_mask: dummy
        ntriplet (int): number of different hyperparameter sets to associate with 3-body pairings
        triplet_mask (np.ndarray): nspec^3 long integer array
        ncut3b (int): number of different 3-body cutoff sets to associate with 3-body pairings
        cut3b_mask (np.ndarray): nspec^2 long integer array
        sig2: dummy
        ls2: dummy
        sig3 (np.ndarray): signal variances associates with three-body term
        ls3 (np.ndarray): length scales associates with three-body term
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        float: Value of the 2+3-body energy/energy kernel.
    """

    if (ncut3b == 0):
        tbmcj = three_body_mc_en_jit
    else:
        tbmcj = three_body_mc_en_sepcut_jit

    return tbmcj(env1.bond_array_3, env1.ctype, env1.etypes,
                 env2.bond_array_3, env2.ctype, env2.etypes,
                 env1.cross_bond_inds, env2.cross_bond_inds,
                 env1.cross_bond_dists, env2.cross_bond_dists,
                 env1.triplet_counts, env2.triplet_counts,
                 sig3, ls3, cutoff_3b, cutoff_func,
                 nspec, spec_mask,
                 triplet_mask, cut3b_mask)/9


# -----------------------------------------------------------------------------
#                       two body multicomponent kernel
# -----------------------------------------------------------------------------


def two_body_mc(
        env1, env2, d1, d2, cutoff_2b, cutoff_3b,  cutoff_mb,  nspec, spec_mask,
        nbond, bond_mask, ntriplet, triplet_mask, ncut3b, cut3b_mask,
        nmb, mb_mask, sig2, ls2, sig3, ls3, sigm, lsm,
        cutoff_func=cf.quadratic_cutoff):
    """2-body multi-element kernel between two force components.

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b (float, np.ndarray): cutoff(s) for two-body interaction
        cutoff_3b: dummy
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond (int): number of different hyperparameter sets to associate with two-body pairings
        bond_mask (np.ndarray): nspec^2 long integer array
        ntriplet: dummy
        triplet_mask: dummy
        ncut3b: dummy
        cut3b_mask: dummy
        sig2: dummy
        ls2: dummy
        sig3: dummy
        ls3: dummy
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        float: Value of the 2-body force/force kernel.
    """

    return two_body_mc_jit(env1.bond_array_2, env1.ctype, env1.etypes,
                           env2.bond_array_2, env2.ctype, env2.etypes,
                           d1, d2, sig2, ls2, cutoff_2b, cutoff_func,
                           nspec, spec_mask, bond_mask)


def two_body_mc_grad(
        env1, env2, d1, d2, cutoff_2b, cutoff_3b,  cutoff_mb,  nspec, spec_mask,
        nbond, bond_mask, ntriplet, triplet_mask,
        ncut3b, cut3b_mask, nmb, mb_mask,
        sig2, ls2, sig3, ls3, sigm, lsm,
        cutoff_func=cf.quadratic_cutoff):
    """2-body multi-element kernel between two force components and its
    gradient with respect to the hyperparameters.

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b (float, np.ndarray): cutoff(s) for two-body interaction
        cutoff_3b: dummy
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond (int): number of different hyperparameter sets to associate with two-body pairings
        bond_mask (np.ndarray): nspec^2 long integer array
        ntriplet: dummy
        triplet_mask: dummy
        ncut3b: dummy
        cut3b_mask: dummy
        sig2 (np.ndarray): signal variances associates with two-body term
        ls2 (np.ndarray): length scales associates with two-body term
        sig3: dummy
        ls3: dummy
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        (float, np.ndarray):
            Value of the 2-body kernel and its gradient
            with respect to the hyperparameters.
    """

    return two_body_mc_grad_jit(
        env1.bond_array_2, env1.ctype, env1.etypes,
        env2.bond_array_2, env2.ctype, env2.etypes,
        d1, d2, sig2, ls2, cutoff_2b, cutoff_func,
        nspec, spec_mask, nbond, bond_mask)


def two_body_mc_force_en(env1, env2, d1, cutoff_2b, cutoff_3b, cutoff_mb,
                         nspec, spec_mask, nbond, bond_mask, ntriplet, triplet_mask,
                         ncut3b, cut3b_mask, nmb, mb_mask,
                         sig2, ls2, sig3, ls3, sigm, lsm,
                         cutoff_func=cf.quadratic_cutoff):
    """2-body multi-element kernel between a force components and local energy

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b (float, np.ndarray): cutoff(s) for two-body interaction
        cutoff_3b: dummy
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond (int): number of different hyperparameter sets to associate with two-body pairings
        bond_mask (np.ndarray): nspec^2 long integer array
        ntriplet: dummy
        triplet_mask: dummy
        ncut3b: dummy
        cut3b_mask: dummy
        sig2: dummy
        ls2: dummy
        sig3: dummy
        ls3: dummy
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        float: Value of the 2-body force/energy kernel.
    """

    return two_body_mc_force_en_jit(
        env1.bond_array_2, env1.ctype, env1.etypes,
        env2.bond_array_2, env2.ctype, env2.etypes,
        d1, sig2, ls2, cutoff_2b, cutoff_func,
        nspec, spec_mask, bond_mask) / 2


def two_body_mc_en(env1, env2, cutoff_2b, cutoff_3b, cutoff_mb,
                   nspec, spec_mask,
                   nbond, bond_mask, ntriplet, triplet_mask,
                   ncut3b, cut3b_mask, nmb, mb_mask,
                   sig2, ls2, sig3, ls3, sigm, lsm,
                   cutoff_func=cf.quadratic_cutoff):
    """2-body multi-element kernel between two local energies

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b (float, np.ndarray): cutoff(s) for two-body interaction
        cutoff_3b: dummy
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond (int): number of different hyperparameter sets to associate with two-body pairings
        bond_mask (np.ndarray): nspec^2 long integer array
        ntriplet: dummy
        triplet_mask: dummy
        ncut3b: dummy
        cut3b_mask: dummy
        sig2: dummy
        ls2: dummy
        sig3: dummy
        ls3: dummy
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        float: Value of the 2-body energy/energy kernel.
    """

    return two_body_mc_en_jit(env1.bond_array_2, env1.ctype, env1.etypes,
                              env2.bond_array_2, env2.ctype, env2.etypes,
                              sig2, ls2, cutoff_2b, cutoff_func,
                              nspec, spec_mask, bond_mask)/4


# -----------------------------------------------------------------------------
#                 three body multicomponent kernel (numba)
# -----------------------------------------------------------------------------

@njit
def three_body_mc_jit(bond_array_1, c1, etypes1,
                      bond_array_2, c2, etypes2,
                      cross_bond_inds_1, cross_bond_inds_2,
                      cross_bond_dists_1, cross_bond_dists_2,
                      triplets_1, triplets_2,
                      d1, d2, sig, ls, r_cut, cutoff_func,
                      nspec, spec_mask, triplet_mask, cut3b_mask):
    kern = 0

    # pre-compute constants that appear in the inner loop
    sig2 = sig * sig
    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2

    bc1 = spec_mask[c1]
    bc1n = bc1 * nspec * nspec

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        ci1 = bond_array_1[m, d1]
        ei1 = etypes1[m]

        bei1 = spec_mask[ei1]
        bei1n = nspec * bei1

        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ei2 = etypes1[ind1]

            tr_spec = [c1, ei1, ei2]
            c2_ind = tr_spec
            if c2 in tr_spec:
                tr_spec.remove(c2)

                ri2 = bond_array_1[ind1, 0]
                ci2 = bond_array_1[ind1, d1]
                fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)

                bei2 = spec_mask[ei2]

                ri3 = cross_bond_dists_1[m, m + n + 1]
                fi3, _ = cutoff_func(r_cut, ri3, 0)

                fi = fi1 * fi2 * fi3
                fdi = fdi1 * fi2 * fi3 + fi1 * fdi2 * fi3

                ttypei = triplet_mask[bc1n + bei1n + bei2]

                tls1 = ls1[ttypei]
                tls2 = ls2[ttypei]
                tls3 = ls3[ttypei]
                tsig2 = sig2[ttypei]

                for p in range(bond_array_2.shape[0]):
                    ej1 = etypes2[p]
                    tr_spec1 = [tr_spec[0], tr_spec[1]]
                    if ej1 in tr_spec1:
                        tr_spec1.remove(ej1)
                        rj1 = bond_array_2[p, 0]
                        cj1 = bond_array_2[p, d2]
                        fj1, fdj1 = cutoff_func(r_cut, rj1, cj1)

                        for q in range(triplets_2[p]):
                            ind2 = cross_bond_inds_2[p, p + 1 + q]
                            ej2 = etypes2[ind2]
                            if ej2 == tr_spec1[0]:
                                ind2 = cross_bond_inds_2[p, p + 1 + q]
                                rj2 = bond_array_2[ind2, 0]
                                cj2 = bond_array_2[ind2, d2]
                                fj2, fdj2 = cutoff_func(r_cut, rj2, cj2)
                                ej2 = etypes2[ind2]

                                rj3 = cross_bond_dists_2[p, p + 1 + q]
                                fj3, _ = cutoff_func(r_cut, rj3, 0)

                                fj = fj1 * fj2 * fj3
                                fdj = fdj1 * fj2 * fj3 + fj1 * fdj2 * fj3

                                r11 = ri1 - rj1
                                r12 = ri1 - rj2
                                r13 = ri1 - rj3
                                r21 = ri2 - rj1
                                r22 = ri2 - rj2
                                r23 = ri2 - rj3
                                r31 = ri3 - rj1
                                r32 = ri3 - rj2
                                r33 = ri3 - rj3

                                if (c1 == c2):
                                    if (ei1 == ej1) and (ei2 == ej2):
                                        kern += \
                                            three_body_helper_1(ci1, ci2, cj1, cj2, r11,
                                                                r22, r33, fi, fj, fdi, fdj,
                                                                tls1, tls2, tls3,
                                                                tsig2)
                                    if (ei1 == ej2) and (ei2 == ej1):
                                        kern += \
                                            three_body_helper_1(ci1, ci2, cj2, cj1, r12,
                                                                r21, r33, fi, fj, fdi, fdj,
                                                                tls1, tls2, tls3,
                                                                tsig2)
                                if (c1 == ej1):
                                    if (ei1 == ej2) and (ei2 == c2):
                                        kern += \
                                            three_body_helper_2(ci2, ci1, cj2, cj1, r21,
                                                                r13, r32, fi, fj, fdi,
                                                                fdj,
                                                                tls1, tls2, tls3,
                                                                tsig2)
                                    if (ei1 == c2) and (ei2 == ej2):
                                        kern += \
                                            three_body_helper_2(ci1, ci2, cj2, cj1, r11,
                                                                r23, r32, fi, fj, fdi,
                                                                fdj,
                                                                tls1, tls2, tls3,
                                                                tsig2)
                                if (c1 == ej2):
                                    if (ei1 == ej1) and (ei2 == c2):
                                        kern += \
                                            three_body_helper_2(ci2, ci1, cj1, cj2, r22,
                                                                r13, r31, fi, fj, fdi,
                                                                fdj,
                                                                tls1, tls2, tls3,
                                                                tsig2)
                                    if (ei1 == c2) and (ei2 == ej1):
                                        kern += \
                                            three_body_helper_2(ci1, ci2, cj1, cj2, r12,
                                                                r23, r31, fi, fj, fdi,
                                                                fdj,
                                                                tls1, tls2, tls3,
                                                                tsig2)

    return kern


@njit
def three_body_mc_grad_jit(bond_array_1, c1, etypes1,
                           bond_array_2, c2, etypes2,
                           cross_bond_inds_1, cross_bond_inds_2,
                           cross_bond_dists_1, cross_bond_dists_2,
                           triplets_1, triplets_2,
                           d1, d2, sig, ls, r_cut, cutoff_func,
                           nspec, spec_mask, ntriplet, triplet_mask,
                           cut3b_mask):
    """Kernel gradient for 3-body force comparisons."""

    kern = 0
    sig_derv = np.zeros(ntriplet, dtype=np.float64)
    ls_derv = np.zeros(ntriplet, dtype=np.float64)

    # pre-compute constants that appear in the inner loop
    sig2, sig3, ls1, ls2, ls3, ls4, ls5, ls6 = grad_constants(sig, ls)

    bc1 = spec_mask[c1]
    bc1n = bc1 * nspec * nspec

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        ci1 = bond_array_1[m, d1]
        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)
        ei1 = etypes1[m]

        bei1 = spec_mask[ei1]
        bei1n = bei1 * nspec

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ei2 = etypes1[ind1]
            tr_spec = [c1, ei1, ei2]
            c2_ind = tr_spec
            if c2 in tr_spec:
                tr_spec.remove(c2)

                ri3 = cross_bond_dists_1[m, m + n + 1]
                ri2 = bond_array_1[ind1, 0]
                ci2 = bond_array_1[ind1, d1]

                bei2 = spec_mask[ei2]

                ttypei = triplet_mask[bc1n + bei1n + bei2]

                tls1 = ls1[ttypei]
                tls2 = ls2[ttypei]
                tls3 = ls3[ttypei]
                tls4 = ls4[ttypei]
                tls5 = ls5[ttypei]
                tls6 = ls6[ttypei]
                tsig2 = sig2[ttypei]
                tsig3 = sig3[ttypei]

                fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)
                fi3, _ = cutoff_func(r_cut, ri3, 0)

                fi = fi1 * fi2 * fi3
                fdi = fdi1 * fi2 * fi3 + fi1 * fdi2 * fi3

                for p in range(bond_array_2.shape[0]):
                    ej1 = etypes2[p]
                    tr_spec1 = [tr_spec[0], tr_spec[1]]
                    if ej1 in tr_spec1:
                        tr_spec1.remove(ej1)
                        rj1 = bond_array_2[p, 0]
                        cj1 = bond_array_2[p, d2]
                        fj1, fdj1 = cutoff_func(r_cut, rj1, cj1)

                        for q in range(triplets_2[p]):
                            ind2 = cross_bond_inds_2[p, p + 1 + q]
                            ej2 = etypes2[ind2]
                            if ej2 == tr_spec1[0]:
                                ind2 = cross_bond_inds_2[p, p + q + 1]
                                rj3 = cross_bond_dists_2[p, p + q + 1]
                                rj2 = bond_array_2[ind2, 0]
                                cj2 = bond_array_2[ind2, d2]
                                ej2 = etypes2[ind2]

                                fj2, fdj2 = cutoff_func(r_cut, rj2, cj2)
                                fj3, _ = cutoff_func(r_cut, rj3, 0)

                                fj = fj1 * fj2 * fj3
                                fdj = fdj1 * fj2 * fj3 + fj1 * fdj2 * fj3

                                r11 = ri1 - rj1
                                r12 = ri1 - rj2
                                r13 = ri1 - rj3
                                r21 = ri2 - rj1
                                r22 = ri2 - rj2
                                r23 = ri2 - rj3
                                r31 = ri3 - rj1
                                r32 = ri3 - rj2
                                r33 = ri3 - rj3

                                if (c1 == c2):
                                    if (ei1 == ej1) and (ei2 == ej2):
                                        kern_term, sig_term, ls_term = \
                                            three_body_grad_helper_1(ci1, ci2, cj1, cj2,
                                                                     r11, r22, r33, fi, fj,
                                                                     fdi, fdj, tls1, tls2,
                                                                     tls3, tls4, tls5,
                                                                     tls6,
                                                                     tsig2, tsig3)
                                        kern += kern_term
                                        sig_derv[ttypei] += sig_term
                                        ls_derv[ttypei] += ls_term

                                    if (ei1 == ej2) and (ei2 == ej1):
                                        kern_term, sig_term, ls_term = \
                                            three_body_grad_helper_1(ci1, ci2, cj2, cj1,
                                                                     r12, r21, r33, fi, fj,
                                                                     fdi, fdj, tls1, tls2,
                                                                     tls3, tls4, tls5,
                                                                     tls6,
                                                                     tsig2, tsig3)
                                        kern += kern_term
                                        sig_derv[ttypei] += sig_term
                                        ls_derv[ttypei] += ls_term

                                if (c1 == ej1):
                                    if (ei1 == ej2) and (ei2 == c2):
                                        kern_term, sig_term, ls_term = \
                                            three_body_grad_helper_2(ci2, ci1, cj2, cj1,
                                                                     r21, r13, r32, fi, fj,
                                                                     fdi, fdj, tls1, tls2,
                                                                     tls3, tls4, tls5,
                                                                     tls6,
                                                                     tsig2, tsig3)
                                        kern += kern_term
                                        sig_derv[ttypei] += sig_term
                                        ls_derv[ttypei] += ls_term

                                    if (ei1 == c2) and (ei2 == ej2):
                                        kern_term, sig_term, ls_term = \
                                            three_body_grad_helper_2(ci1, ci2, cj2, cj1,
                                                                     r11, r23, r32, fi, fj,
                                                                     fdi, fdj, tls1, tls2,
                                                                     tls3, tls4, tls5,
                                                                     tls6,
                                                                     tsig2, tsig3)
                                        kern += kern_term
                                        sig_derv[ttypei] += sig_term
                                        ls_derv[ttypei] += ls_term

                                if (c1 == ej2):
                                    if (ei1 == ej1) and (ei2 == c2):
                                        kern_term, sig_term, ls_term = \
                                            three_body_grad_helper_2(ci2, ci1, cj1, cj2,
                                                                     r22, r13, r31, fi, fj,
                                                                     fdi, fdj, tls1, tls2,
                                                                     tls3, tls4, tls5,
                                                                     tls6,
                                                                     tsig2, tsig3)
                                        kern += kern_term
                                        sig_derv[ttypei] += sig_term
                                        ls_derv[ttypei] += ls_term

                                    if (ei1 == c2) and (ei2 == ej1):
                                        kern_term, sig_term, ls_term = \
                                            three_body_grad_helper_2(ci1, ci2, cj1, cj2,
                                                                     r12, r23, r31, fi, fj,
                                                                     fdi, fdj, tls1, tls2,
                                                                     tls3, tls4, tls5,
                                                                     tls6,
                                                                     tsig2, tsig3)

                                        kern += kern_term
                                        sig_derv[ttypei] += sig_term
                                        ls_derv[ttypei] += ls_term

    return kern, np.hstack((sig_derv, ls_derv))


@njit
def three_body_mc_force_en_jit(bond_array_1, c1, etypes1,
                               bond_array_2, c2, etypes2,
                               cross_bond_inds_1, cross_bond_inds_2,
                               cross_bond_dists_1, cross_bond_dists_2,
                               triplets_1, triplets_2,
                               d1, sig, ls, r_cut, cutoff_func,
                               nspec, spec_mask, triplet_mask, cut3b_mask):
    """Kernel for 3-body force/energy comparisons."""

    kern = 0

    # pre-compute constants that appear in the inner loop
    sig2 = sig * sig
    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)

    bc1 = spec_mask[c1]
    bc1n = nspec * nspec * bc1

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        ci1 = bond_array_1[m, d1]
        fi1, fdi1 = cutoff_func(r_cut, ri1, ci1)
        ei1 = etypes1[m]

        bei1 = spec_mask[ei1]
        bei1n = nspec * bei1

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ei2 = etypes1[ind1]
            tr_spec = [c1, ei1, ei2]
            c2_ind = tr_spec
            if c2 in tr_spec:
                tr_spec.remove(c2)
                ri2 = bond_array_1[ind1, 0]
                ci2 = bond_array_1[ind1, d1]
                fi2, fdi2 = cutoff_func(r_cut, ri2, ci2)

                bei2 = spec_mask[ei2]

                ttypei = triplet_mask[bc1n + bei1n + bei2]

                tls1 = ls1[ttypei]
                tls2 = ls2[ttypei]
                tsig2 = sig2[ttypei]

                ri3 = cross_bond_dists_1[m, m + n + 1]
                fi3, _ = cutoff_func(r_cut, ri3, 0)

                fi = fi1 * fi2 * fi3
                fdi = fdi1 * fi2 * fi3 + fi1 * fdi2 * fi3

                for p in range(bond_array_2.shape[0]):

                    ej1 = etypes2[p]
                    tr_spec1 = [tr_spec[0], tr_spec[1]]
                    if ej1 in tr_spec1:
                        tr_spec1.remove(ej1)
                        rj1 = bond_array_2[p, 0]
                        fj1, _ = cutoff_func(r_cut, rj1, 0)

                        for q in range(triplets_2[p]):

                            ind2 = cross_bond_inds_2[p, p + 1 + q]
                            ej2 = etypes2[ind2]
                            if ej2 == tr_spec1[0]:

                                ind2 = cross_bond_inds_2[p, p + q + 1]
                                rj2 = bond_array_2[ind2, 0]
                                fj2, _ = cutoff_func(r_cut, rj2, 0)
                                ej2 = etypes2[ind2]
                                rj3 = cross_bond_dists_2[p, p + q + 1]
                                fj3, _ = cutoff_func(r_cut, rj3, 0)
                                fj = fj1 * fj2 * fj3

                                r11 = ri1 - rj1
                                r12 = ri1 - rj2
                                r13 = ri1 - rj3
                                r21 = ri2 - rj1
                                r22 = ri2 - rj2
                                r23 = ri2 - rj3
                                r31 = ri3 - rj1
                                r32 = ri3 - rj2
                                r33 = ri3 - rj3

                                if (c1 == c2):
                                    if (ei1 == ej1) and (ei2 == ej2):
                                        kern += three_body_en_helper(ci1, ci2, r11, r22,
                                                                     r33, fi, fj, fdi,
                                                                     tls1,
                                                                     tls2, tsig2)
                                    if (ei1 == ej2) and (ei2 == ej1):
                                        kern += three_body_en_helper(ci1, ci2, r12, r21,
                                                                     r33, fi, fj, fdi,
                                                                     tls1,
                                                                     tls2, tsig2)
                                if (c1 == ej1):
                                    if (ei1 == ej2) and (ei2 == c2):
                                        kern += three_body_en_helper(ci1, ci2, r13, r21,
                                                                     r32, fi, fj, fdi,
                                                                     tls1,
                                                                     tls2, tsig2)
                                    if (ei1 == c2) and (ei2 == ej2):
                                        kern += three_body_en_helper(ci1, ci2, r11, r23,
                                                                     r32, fi, fj, fdi,
                                                                     tls1,
                                                                     tls2, tsig2)
                                if (c1 == ej2):
                                    if (ei1 == ej1) and (ei2 == c2):
                                        kern += three_body_en_helper(ci1, ci2, r13, r22,
                                                                     r31, fi, fj, fdi,
                                                                     tls1,
                                                                     tls2, tsig2)
                                    if (ei1 == c2) and (ei2 == ej1):
                                        kern += three_body_en_helper(ci1, ci2, r12, r23,
                                                                     r31, fi, fj, fdi,
                                                                     tls1,
                                                                     tls2, tsig2)

    return kern


@njit
def three_body_mc_en_jit(bond_array_1, c1, etypes1,
                         bond_array_2, c2, etypes2,
                         cross_bond_inds_1, cross_bond_inds_2,
                         cross_bond_dists_1, cross_bond_dists_2,
                         triplets_1, triplets_2,
                         sig, ls, r_cut, cutoff_func,
                         nspec, spec_mask, triplet_mask, cut3b_mask):
    kern = 0

    sig2 = sig * sig
    ls2 = 1 / (2 * ls * ls)

    bc1 = spec_mask[c1]
    bc1n = bc1 * nspec * nspec

    for m in range(bond_array_1.shape[0]):
        ri1 = bond_array_1[m, 0]
        fi1, _ = cutoff_func(r_cut, ri1, 0)
        ei1 = etypes1[m]

        bei1 = spec_mask[ei1]
        bei1n = nspec * bei1

        for n in range(triplets_1[m]):
            ind1 = cross_bond_inds_1[m, m + n + 1]
            ei2 = etypes1[ind1]
            tr_spec = [c1, ei1, ei2]
            c2_ind = tr_spec
            if c2 in tr_spec:
                tr_spec.remove(c2)
                ri2 = bond_array_1[ind1, 0]
                fi2, _ = cutoff_func(r_cut, ri2, 0)

                bei2 = spec_mask[ei2]

                ttypei = triplet_mask[bc1n + bei1n + bei2]

                tls2 = ls2[ttypei]
                tsig2 = sig2[ttypei]

                ri3 = cross_bond_dists_1[m, m + n + 1]
                fi3, _ = cutoff_func(r_cut, ri3, 0)
                fi = fi1 * fi2 * fi3

                for p in range(bond_array_2.shape[0]):
                    ej1 = etypes2[p]
                    tr_spec1 = [tr_spec[0], tr_spec[1]]
                    if ej1 in tr_spec1:
                        tr_spec1.remove(ej1)
                        rj1 = bond_array_2[p, 0]
                        fj1, _ = cutoff_func(r_cut, rj1, 0)

                        for q in range(triplets_2[p]):
                            ind2 = cross_bond_inds_2[p, p + 1 + q]
                            ej2 = etypes2[ind2]
                            if ej2 == tr_spec1[0]:
                                rj2 = bond_array_2[ind2, 0]
                                fj2, _ = cutoff_func(r_cut, rj2, 0)

                                rj3 = cross_bond_dists_2[p, p + q + 1]
                                fj3, _ = cutoff_func(r_cut, rj3, 0)
                                fj = fj1 * fj2 * fj3

                                r11 = ri1 - rj1
                                r12 = ri1 - rj2
                                r13 = ri1 - rj3
                                r21 = ri2 - rj1
                                r22 = ri2 - rj2
                                r23 = ri2 - rj3
                                r31 = ri3 - rj1
                                r32 = ri3 - rj2
                                r33 = ri3 - rj3

                                if (c1 == c2):
                                    if (ei1 == ej1) and (ei2 == ej2):
                                        C1 = r11 * r11 + r22 * r22 + r33 * r33
                                        kern += tsig2 * \
                                            exp(-C1 * tls2) * fi * fj
                                    if (ei1 == ej2) and (ei2 == ej1):
                                        C3 = r12 * r12 + r21 * r21 + r33 * r33
                                        kern += tsig2 * \
                                            exp(-C3 * tls2) * fi * fj
                                if (c1 == ej1):
                                    if (ei1 == ej2) and (ei2 == c2):
                                        C5 = r13 * r13 + r21 * r21 + r32 * r32
                                        kern += tsig2 * \
                                            exp(-C5 * tls2) * fi * fj
                                    if (ei1 == c2) and (ei2 == ej2):
                                        C2 = r11 * r11 + r23 * r23 + r32 * r32
                                        kern += tsig2 * \
                                            exp(-C2 * tls2) * fi * fj
                                if (c1 == ej2):
                                    if (ei1 == ej1) and (ei2 == c2):
                                        C6 = r13 * r13 + r22 * r22 + r31 * r31
                                        kern += tsig2 * \
                                            exp(-C6 * tls2) * fi * fj
                                    if (ei1 == c2) and (ei2 == ej1):
                                        C4 = r12 * r12 + r23 * r23 + r31 * r31
                                        kern += tsig2 * \
                                            exp(-C4 * tls2) * fi * fj

    return kern


# -----------------------------------------------------------------------------
#                 two body multicomponent kernel (numba)
# -----------------------------------------------------------------------------


@njit
def two_body_mc_jit(bond_array_1, c1, etypes1, bond_array_2, c2, etypes2,
                    d1, d2, sig, ls, r_cut, cutoff_func,
                    nspec, spec_mask, bond_mask):
    """Multicomponent two-body force/force kernel accelerated with Numba's
    njit decorator.
    Loops over bonds in two environments and adds to the kernel if bonds are
    of the same type.
    """

    kern = 0

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2
    sig2 = sig * sig

    bc1 = spec_mask[c1]
    bc1n = bc1 * nspec

    for m in range(bond_array_1.shape[0]):
        e1 = etypes1[m]

        if ((c2 == e1) or (c2 == c1)):
            ri = bond_array_1[m, 0]
            ci = bond_array_1[m, d1]

            be1 = spec_mask[e1]
            btype = bond_mask[bc1n + be1]

            tls1 = ls1[btype]
            tls2 = ls2[btype]
            tls3 = ls3[btype]
            tsig2 = sig2[btype]
            tr_cut = r_cut[btype]

            fi, fdi = cutoff_func(tr_cut, ri, ci)

            for n in range(bond_array_2.shape[0]):
                e2 = etypes2[n]

                # check if bonds agree
                if (c1 == c2 and e1 == e2) or (c1 == e2 and c2 == e1):
                    rj = bond_array_2[n, 0]
                    cj = bond_array_2[n, d2]
                    fj, fdj = cutoff_func(tr_cut, rj, cj)
                    r11 = ri - rj

                    A = ci * cj
                    B = r11 * ci
                    C = r11 * cj
                    D = r11 * r11

                    kern += force_helper(A, B, C, D, fi, fj, fdi, fdj,
                                         tls1, tls2, tls3, tsig2)

    return kern


@njit
def two_body_mc_grad_jit(bond_array_1, c1, etypes1,
                         bond_array_2, c2, etypes2,
                         d1, d2, sig, ls, r_cut, cutoff_func,
                         nspec, spec_mask, nbond, bond_mask):
    """Multicomponent two-body force/force kernel gradient accelerated with
    Numba's njit decorator."""

    kern = 0
    sig_derv = np.zeros(nbond, dtype=np.float64)
    ls_derv = np.zeros(nbond, dtype=np.float64)

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    ls3 = ls2 * ls2
    ls4 = 1 / (ls * ls * ls)
    ls5 = ls * ls
    ls6 = ls2 * ls4

    sig2 = sig * sig
    sig3 = 2 * sig

    bc1 = spec_mask[c1]
    bc1n = bc1 * nspec

    for m in range(bond_array_1.shape[0]):
        e1 = etypes1[m]
        if ((c2 == e1) or (c2 == c1)):
            ri = bond_array_1[m, 0]
            ci = bond_array_1[m, d1]

            be1 = spec_mask[e1]
            btype = bond_mask[bc1n + be1]

            tls1 = ls1[btype]
            tls2 = ls2[btype]
            tls3 = ls3[btype]
            tls4 = ls4[btype]
            tls5 = ls5[btype]
            tls6 = ls6[btype]
            tsig2 = sig2[btype]
            tsig3 = sig3[btype]
            tr_cut = r_cut[btype]

            fi, fdi = cutoff_func(tr_cut, ri, ci)

            for n in range(bond_array_2.shape[0]):
                e2 = etypes2[n]

                # check if bonds agree
                if (c1 == c2 and e1 == e2) or (c1 == e2 and c2 == e1):
                    rj = bond_array_2[n, 0]
                    cj = bond_array_2[n, d2]
                    fj, fdj = cutoff_func(tr_cut, rj, cj)

                    r11 = ri - rj

                    A = ci * cj
                    B = r11 * ci
                    C = r11 * cj
                    D = r11 * r11

                    kern_term, sig_term, ls_term = \
                        grad_helper(A, B, C, D, fi, fj, fdi, fdj,
                                    tls1, tls2, tls3,
                                    tls4, tls5, tls6,
                                    tsig2, tsig3)

                    kern += kern_term
                    sig_derv[btype] += sig_term
                    ls_derv[btype] += ls_term

    kern_grad = np.hstack((sig_derv, ls_derv))

    return kern, kern_grad


@njit
def two_body_mc_force_en_jit(bond_array_1, c1, etypes1,
                             bond_array_2, c2, etypes2,
                             d1, sig, ls, r_cut, cutoff_func,
                             nspec, spec_mask, bond_mask):
    """Multicomponent two-body force/energy kernel accelerated with
    Numba's njit decorator."""

    kern = 0

    ls1 = 1 / (2 * ls * ls)
    ls2 = 1 / (ls * ls)
    sig2 = sig * sig

    bc1 = spec_mask[c1]
    bc1n = bc1 * nspec

    for m in range(bond_array_1.shape[0]):
        e1 = etypes1[m]
        if ((c2 == e1) or (c2 == c1)):
            ri = bond_array_1[m, 0]
            ci = bond_array_1[m, d1]

            be1 = spec_mask[e1]
            btype = bond_mask[bc1n + be1]

            tls1 = ls1[btype]
            tls2 = ls2[btype]
            tsig2 = sig2[btype]
            tr_cut = r_cut[btype]

            fi, fdi = cutoff_func(tr_cut, ri, ci)

            for n in range(bond_array_2.shape[0]):
                e2 = etypes2[n]

                # check if bonds agree
                if (c1 == c2 and e1 == e2) or (c1 == e2 and c2 == e1):
                    rj = bond_array_2[n, 0]
                    fj, _ = cutoff_func(tr_cut, rj, 0)

                    r11 = ri - rj
                    B = r11 * ci
                    D = r11 * r11
                    kern += force_energy_helper(B, D, fi, fj, fdi,
                                                tls1, tls2,
                                                tsig2)

    return kern


@njit
def two_body_mc_en_jit(bond_array_1, c1, etypes1,
                       bond_array_2, c2, etypes2,
                       sig, ls, r_cut, cutoff_func,
                       nspec, spec_mask, bond_mask):
    """Multicomponent two-body energy/energy kernel accelerated with
    Numba's njit decorator."""

    kern = 0

    ls1 = 1 / (2 * ls * ls)
    sig2 = sig * sig

    bc1 = spec_mask[c1]
    bc1n = bc1 * nspec

    for m in range(bond_array_1.shape[0]):
        e1 = etypes1[m]
        if ((c2 == e1) or (c2 == c1)):
            ri = bond_array_1[m, 0]

            be1 = spec_mask[e1]
            btype = bond_mask[bc1n + be1]

            tls1 = ls1[btype]
            tsig2 = sig2[btype]
            tr_cut = r_cut[btype]
            fi, _ = cutoff_func(tr_cut, ri, 0)

            for n in range(bond_array_2.shape[0]):
                e2 = etypes2[n]

                if (c1 == c2 and e1 == e2) or (c1 == e2 and c2 == e1):
                    rj = bond_array_2[n, 0]
                    fj, _ = cutoff_func(tr_cut, rj, 0)
                    r11 = ri - rj
                    kern += fi * fj * tsig2 * exp(-r11 * r11 * tls1)

    return kern


def many_body_mc(env1, env2, d1, d2, cutoff_2b, cutoff_3b, cutoff_mb,
                 nspec, spec_mask,
                 nbond, bond_mask, ntriplet, triplet_mask,
                 ncut3b, cut3b_mask,
                 nmb, mb_mask,
                 sig2, ls2, sig3, ls3, sigm, lsm,
                 cutoff_func=cf.quadratic_cutoff):
    """many-body multi-element kernel between two force components.

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b: dummy
        cutoff_3b: dummy
        cutoff_mb (float, np.ndarray): cutoff(s) for coordination-based manybody interaction
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond: dummy
        bond_mask: dummy
        ntriplet: dummy
        triplet_mask: dummy
        ncut3b: dummy
        cut3b_mask: dummy
        nmb (int): number of different hyperparameter sets to associate with manybody pairings
        mb_mask (np.ndarray): nspec^2 long integer array
        sig2: dummy
        ls2: dummy
        sig3: dummy
        ls3: dummy
        sigm (np.ndarray): signal variances associates with manybody term
        lsm (np.ndarray): length scales associates with manybody term
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        float: Value of the 2+3+many-body kernel.
    """
    return many_body_mc_sepcut_jit(env1.q_array, env2.q_array,
                                   env1.q_neigh_array, env2.q_neigh_array,
                                   env1.q_neigh_grads, env2.q_neigh_grads,
                                   env1.ctype, env2.ctype,
                                   env1.etypes_mb, env2.etypes_mb,
                                   env1.unique_species, env2.unique_species,
                                   d1, d2, sigm, lsm,
                                   nspec, spec_mask, mb_mask)


def many_body_mc_grad(env1, env2, d1, d2, cutoff_2b, cutoff_3b, cutoff_mb,
                      nspec, spec_mask,
                      nbond, bond_mask, ntriplet, triplet_mask,
                      ncut3b, cut3b_mask,
                      nmb, mb_mask,
                      sig2, ls2, sig3, ls3, sigm, lsm,
                      cutoff_func=cf.quadratic_cutoff):
    """manybody multi-element kernel between two force components and its
    gradient with respect to the hyperparameters.

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        cutoff_2b: dummy
        cutoff_3b: dummy
        cutoff_mb (float, np.ndarray): cutoff(s) for coordination-based manybody interaction
        nspec (int): number of different species groups
        spec_mask (np.ndarray): 118-long integer array that determines specie group
        nbond: dummy
        bond_mask: dummy
        ntriplet: dummy
        triplet_mask: dummy
        ncut3b: dummy
        cut3b_mask: dummy
        nmb (int): number of different hyperparameter sets to associate with manybody pairings
        mb_mask (np.ndarray): nspec^2 long integer array
        sig2: dummy
        ls2: dummy
        sig3: dummy
        ls3: dummy
        sigm (np.ndarray): signal variances associates with manybody term
        lsm (np.ndarray): length scales associates with manybody term
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        (float, np.ndarray):
            Value of the 2+3+manybody kernel and its gradient
            with respect to the hyperparameters.
    """

    return many_body_mc_grad_sepcut_jit(env1.q_array, env2.q_array,
                                        env1.q_neigh_array, env2.q_neigh_array,
                                        env1.q_neigh_grads, env2.q_neigh_grads,
                                        env1.ctype, env2.ctype,
                                        env1.etypes_mb, env2.etypes_mb,
                                        env1.unique_species, env2.unique_species,
                                        d1, d2, sigm, lsm,
                                        nspec, spec_mask, nmb, mb_mask)


def many_body_mc_force_en(env1, env2, d1, cutoff_2b, cutoff_3b, cutoff_mb,
                          nspec, spec_mask,
                          nbond, bond_mask, ntriplet, triplet_mask,
                          ncut3b, cut3b_mask,
                          nmb, mb_mask,
                          sig2, ls2, sig3, ls3, sigm, lsm,
                          cutoff_func=cf.quadratic_cutoff):
    """many-body single-element kernel between two local energies.

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        hyps (np.ndarray): Hyperparameters of the kernel function (sig, ls).
        cutoffs (np.ndarray): Two-element array containing the 2-, 3-, and
            many-body cutoffs.
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        float: Value of the many-body force/energy kernel.
    """

    return many_body_mc_force_en_sepcut_jit(env1.q_array, env2.q_array,
                                            env1.q_neigh_array,
                                            env1.q_neigh_grads,
                                            env1.ctype, env2.ctype,
                                            env1.etypes_mb,
                                            env1.unique_species, env2.unique_species,
                                            d1, sigm, lsm,
                                            nspec, spec_mask, mb_mask)


def many_body_mc_en(env1, env2, cutoff_2b, cutoff_3b, cutoff_mb,
                    nspec, spec_mask,
                    nbond, bond_mask, ntriplet, triplet_mask,
                    ncut3b, cut3b_mask,
                    nmb, mb_mask,
                    sig2, ls2, sig3, ls3, sigm, lsm,
                    cutoff_func=cf.quadratic_cutoff):
    """many-body multi-element kernel between two local energies.

    Args:
        env1 (AtomicEnvironment): First local environment.
        env2 (AtomicEnvironment): Second local environment.
        hyps (np.ndarray): Hyperparameters of the kernel function (sig, ls).
        cutoffs (np.ndarray): One-element array containing the 2-body
            cutoff.
        cutoff_func (Callable): Cutoff function of the kernel.

    Return:
        float: Value of the 2-body force/energy kernel.
    """

    return many_body_mc_en_sepcut_jit(env1.q_array, env2.q_array,
                                      env1.ctype, env2.ctype,
                                      env1.unique_species, env2.unique_species,
                                      sigm, lsm,
                                      nspec, spec_mask, mb_mask)


_str_to_kernel = {'2': two_body_mc,
                  '2_en': two_body_mc_en,
                  '2_grad': two_body_mc_grad,
                  '2_force_en': two_body_mc_force_en,
                  '2_efs_energy': 'not implemented',
                  '2_efs_force': 'not implemented',
                  '2_efs_self': 'not implemented',
                  '3': three_body_mc,
                  '3_grad': three_body_mc_grad,
                  '3_en': three_body_mc_en,
                  '3_force_en': three_body_mc_force_en,
                  '3_efs_energy': 'not implemented',
                  '3_efs_force': 'not implemented',
                  '3_efs_self': 'not implemented',
                  'many': many_body_mc,
                  'many_grad': many_body_mc_grad,
                  'many_en': many_body_mc_en,
                  'many_force_en': many_body_mc_force_en,
                  'many_efs_energy': 'not implemented',
                  'many_efs_force': 'not implemented',
                  'many_efs_self': 'not implemented',
                  '2+3': two_plus_three_body_mc,
                  '2+3_grad': two_plus_three_body_mc_grad,
                  '2+3_en': two_plus_three_mc_en,
                  '2+3_force_en': two_plus_three_mc_force_en,
                  '2+3_efs_energy': 'not implemented',
                  '2+3_efs_force': 'not implemented',
                  '2+3_efs_self': 'not implemented',
                  '2+3+many': two_three_many_body_mc,
                  '2+3+many_grad': two_three_many_body_mc_grad,
                  '2+3+many_en': two_three_many_mc_en,
                  '2+3+many_force_en': two_three_many_mc_force_en,
                  '2+3+many_efs_energy': 'not implemented',
                  '2+3+many_efs_force': 'not implemented',
                  '2+3+many_efs_self': 'not implemented'
                  }



"""
:class:`MappedGaussianProcess` uses splines to build up interpolation\
function of the low-dimensional decomposition of Gaussian Process, \
with little loss of accuracy. Refer to \
`Vandermause et al. <https://www.nature.com/articles/s41524-020-0283-z>`_, \
`Glielmo et al. <https://journals.aps.org/prb/abstract/10.1103/PhysRevB.97.184307>`_
"""
import time, os, math, inspect, subprocess, json, warnings, pickle
import numpy as np
import multiprocessing as mp

from copy import deepcopy
from typing import List

from flare.env import AtomicEnvironment
from flare.gp import GaussianProcess
from flare.utils.element_coder import NumpyEncoder, element_to_Z, Z_to_element

from flare.mgp.map2b import Map2body
from flare.mgp.map3b import Map3body


class MappedGaussianProcess:
    """
    Build Mapped Gaussian Process (MGP)
    and automatically save coefficients for LAMMPS pair style.

    Args:
        grid_params (dict): Parameters for the mapping itself, such as
            grid size of spline fit, etc. As described below.
        unique_species (dict): List of all the (unique) species included during
            the training that need to be mapped
        GP (GaussianProcess): None or a GaussianProcess object. If a GP is input,
            and container_only is False, automatically build a mapping corresponding
            to the GaussianProcess.
        var_map (str): if None: only build mapping for mean (force). If 'pca', then 
            use PCA to map the variance, based on `grid_params['xxbody']['svd_rank']`.
            If 'simple', then only map the diagonal of covariance, and predict the 
            upper bound of variance. The 'pca' mode is much heavier in terms of 
            memory, but its prediction is much closer to GP variance.
        container_only (bool): if True: only build splines container
            (with no coefficients); if False: Attempt to build map immediately
        lmp_file_name (str): LAMMPS coefficient file name
        n_cpus (int): Default None. Set to the number of cores needed for
            parallelization. Used in the construction of the map.
        n_sample (int): Default 10. The batch size for building map. Not used now.

    Examples:

    >>> # build 2 + 3 body map
    >>> grid_params = {'twobody': {'grid_num': [64]},
    ...                'threebody': {'grid_num': [64, 64, 64]}}

    For `grid_params`, the following keys and values are allowed

    Args:
        'twobody' (dict, optional): if 2-body is present, set as a dictionary
            of parameters for 2-body mapping. Parameters see below.
        'threebody' (dict, optional): if 3-body is present, set as a dictionary
            of parameters for 3-body mapping. Parameters see below.
        'load_grid' (str, optional): Default None. the path to the directory
            where the previously generated grids (``grid_*.npy``) are stored.
            If no path is specified, MGP will construct grids from scratch.
        'lower_bound_relax' (float, optional): Default 0.1. if 'lower_bound' is
            set to 'auto' this value will be used as a relaxation of lower
            bound. (see below the description of 'lower_bound')

    For two/three body parameter dictionary, the following keys and values are allowed

    Args:
        'grid_num' (list): a list of integers, the number of grid points for
            interpolation. The larger the number, the better the approximation
            of MGP is compared with GP.
        'lower_bound' (str or list, optional): Default 'auto', the lower bound
            of the spline interpolation will be searched. First, search the
            training set of GP and find the minimal interatomic distance r_min.
            Then, the ``lower_bound = r_min - lower_bound_relax``. The user
            can set their own lower_bound, of the same shape as 'grid_num'.
            E.g. for threebody, the customized lower bound can be set as
            [1.2, 1.2, 1.2].
        'upper_bound' (str or list, optional): Default 'auto', the upper bound
            of the spline interpolation will be the cutoffs of GP. The user
            can set their own upper_bound, of the same shape as 'grid_num'.
            E.g. for threebody, the customized lower bound can be set as
            [3.5, 3.5, 3.5].
        'svd_rank' (int, optional): Default 'auto'. If the variance mapping is
            needed, it is set as the rank of the mapping. 'auto' uses full
            rank, which is the smaller one between the total number of grid
            points and training set size. i.e.
            ``full_rank = min(np.prod(grid_num), 3 * N_train)``
    """

    def __init__(
        self,
        grid_params: dict,
        unique_species: list = [],
        GP: GaussianProcess = None,
        var_map: str = None,
        container_only: bool = True,
        lmp_file_name: str = "lmp",
        n_cpus: int = None,
        n_sample: int = 10,
    ):

        # load all arguments as attributes
        self.var_map = var_map
        self.lmp_file_name = lmp_file_name
        self.n_cpus = n_cpus
        self.n_sample = n_sample
        self.grid_params = grid_params
        self.species_labels = []
        self.coded_species = []

        self.hyps_mask = None
        self.cutoffs = None
        self.training_statistics = None

        species_labels = []
        coded_species = []
        for i, ele in enumerate(unique_species):
            if isinstance(ele, str):
                species_labels.append(ele)
                coded_species.append(element_to_Z(ele))
            elif isinstance(ele, int):
                coded_species.append(ele)
                species_labels.append(Z_to_element(ele))
            else:
                print("element type not accepted", ele, type(ele))
        sort_id = np.argsort(coded_species)
        for i in sort_id:
            self.coded_species.append(coded_species[i])
            self.species_labels.append(species_labels[i])

        self.load_grid = grid_params.get("load_grid", None)
        self.update = grid_params.get("update", False)
        self.lower_bound_relax = grid_params.get("lower_bound_relax", 0.1)

        self.maps = {}

        optional_xb_params = ["lower_bound", "upper_bound", "svd_rank"]
        for key in grid_params:
            if "body" in key:
                if "twobody" == key:
                    mapxbody = Map2body
                elif "threebody" == key:
                    mapxbody = Map3body
                else:
                    raise KeyError("Only 'twobody' & 'threebody' are allowed")

                xb_dict = grid_params[key]

                # set to 'auto' if the param is not given
                args = {}
                for oxp in optional_xb_params:
                    args[oxp] = xb_dict.get(oxp, "auto")
                args["grid_num"] = xb_dict.get("grid_num", None)

                for k in xb_dict:
                    args[k] = xb_dict[k]

                xb_maps = mapxbody(**args, **self.__dict__)
                self.maps[key] = xb_maps

    def build_map(self, GP: GaussianProcess):
        self.hyps_mask = GP.hyps_mask
        self.cutoffs = GP.cutoffs
        self.training_statistics = GP.training_statistics

        for xb in self.maps:
            self.maps[xb].build_map(GP)

        # write to lammps pair style coefficient file
        self.write_lmp_file(self.lmp_file_name + '.mgp', write_var=False)
        if self.var_map == 'simple':
            self.write_lmp_file(self.lmp_file_name + '.var', write_var=True)

    def predict(
        self, atom_env: AtomicEnvironment
    ) -> ("ndarray", "ndarray", "ndarray", float):
        """
        predict force, variance, stress and local energy for given
        atomic environment

        Args:
            atom_env: atomic environment (with a center atom and its neighbors)

        Return:
            force: 3d array of atomic force
            variance: 3d array of the predictive variance
            stress: 6d array of the virial stress
            energy: the local energy (atomic energy)
        """

        force = virial = kern = v = energy = 0
        rebuild_dict = {}
        newbound_dict = {}
        for xb in self.maps:
            try:
                pred = self.maps[xb].predict(atom_env)
                force += pred[0]
                virial += pred[1]
                kern += pred[2]
                v += pred[3]
                energy += pred[4]
            # record maps when test envs have smaller distance than lower bound
            except ValueError as err_msg:
                rebuild_dict[xb] = err_msg.args[0]
                newbound_dict[xb] = err_msg.args[1]

        if len(rebuild_dict) > 0:
            err_str = "\n"
            for xb in rebuild_dict:
                nb = len(rebuild_dict[xb])
                err_str += f"{nb} {xb} maps need re-construction\n"
            raise ValueError(rebuild_dict, newbound_dict, err_str)

        if self.var_map == "simple":
            variance = v ** 2
        else:
            variance = kern - np.sum(v ** 2, axis=0)

        return force, variance, virial, energy


    def write_lmp_file(self, lammps_name: str, write_var: bool=False):
        """
        write the coefficients to a file that can be used by lammps pair style
        """

        f = open(lammps_name, "w")

        # write header
        header_comment = """# #2bodyarray #3bodyarray\n# elem1 elem2 a b order\n\n"""
        f.write(header_comment)
        header = ""
        xbodies = ["twobody", "threebody"]
        for xb in xbodies:
            if xb in self.maps:
                num = self.maps[xb].num_lmp_maps #len(self.maps[xb].maps)
            else:
                num = 0
            header += f"{num} "
        f.write(header + "\n")

        # write coefficients
        for xb in self.maps:
            self.maps[xb].write(f, write_var)

        f.close()

    def as_dict(self) -> dict:
        """
        Dictionary representation of the MGP model.
        """

        out_dict = deepcopy(dict(vars(self)))
        out_dict.pop("maps")

        # Uncertainty mappings currently not serializable;
        if self.var_map == 'pca':
            warnings.warn(
                "Uncertainty mappings cannot be serialized, "
                "and so the MGP dict outputted will not have "
                "them.",
                Warning,
            )
            out_dict["var_map"] = None

        # only save the coefficients
        maps_dict = {}
        for m in self.maps:
            maps_dict[m] = self.maps[m].as_dict()
        out_dict["maps"] = maps_dict

        return out_dict

    @staticmethod
    def from_dict(dictionary: dict) -> 'MappedGaussianProcess':
        """
        Create MGP object from dictionary representation.
        """

        # Set GP
        if dictionary.get("GP"):
            GP = GaussianProcess.from_dict(dictionary.get("GP"))
        else:
            dictionary["GP"] = None

        dictionary["unique_species"] = list(set(dictionary["species_labels"]))
        if "container_only" not in dictionary:
            dictionary["container_only"] = True

        init_arg_name = [
            "grid_params",
            "unique_species",
            "GP",
            "var_map",
            "container_only",
            "lmp_file_name",
            "n_cpus",
            "n_sample",
        ]
        kwargs = {key: dictionary[key] for key in init_arg_name}
        new_mgp = MappedGaussianProcess(**kwargs)

        # Fill up the model with the saved coeffs
        if "twobody" in new_mgp.maps:
            new_mgp.maps["twobody"] = Map2body.from_dict(
                dictionary["maps"]["twobody"], Map2body
            )
        if "threebody" in new_mgp.maps:
            new_mgp.maps["threebody"] = Map3body.from_dict(
                dictionary["maps"]["threebody"], Map3body
            )

        return new_mgp

    def write_model(self, name: str, format: str="json"):
        """
        Write everything necessary to re-load and re-use the model
        :param model_name:
        :return:
        """
        if "json" in format.lower() or 'json' in name:
            with open(f"{name}.json", "w") as f:
                json.dump(self.as_dict(), f, cls=NumpyEncoder)

        elif "pickle" in format.lower() or "binary" in format.lower()\
                or 'pickle' in name or 'binary' in name:
            with open(f"{name}.pickle", "wb") as f:
                pickle.dump(self, f)

        else:
            raise ValueError("Requested format not found.")

    @staticmethod
    def from_file(filename: str, format: str = ''):
        if ".json" in filename or format.lower() == 'json':
            with open(filename, "r") as f:
                model = MappedGaussianProcess.from_dict(json.loads(f.readline()))
            return model

        elif "pickle" in filename or format.lower() in ['binary', 'pickle']:
            with open(filename, "rb") as f:
                return pickle.load(f)
        else:
            raise NotImplementedError



import numpy as np

from flare.kernels import sc, mc_simple, mc_sephyps
from flare.parameters import Parameters

"""
This module includes interface functions between kernels and gp/gp_algebra

str_to_kernel_set is used in GaussianProcess class to search for a kernel
    function based on a string name.

from_mask_to_args converts the hyperparameter vector and the dictionary of
    hyps_mask to the list of arguments needed by the kernel function.

from_grad_to_mask(grad, hyps_mask) converts the gradient matrix to the actual
    gradient matrix by removing the fixed dimensions.
"""


def str_to_kernel_set(kernels: list = ['twobody', 'threebody'],
                      component: str = "mc",
                      hyps_mask: dict = None):
    """
    return kernels and kernel gradient function base on a string.
    If it contains 'sc', it will use the kernel in sc module;
    otherwise, it uses the kernel in mc_simple;
    if sc is not included and multihyps is True,
    it will use the kernel in mc_sephyps module
    otherwise, it will use the kernel in the sc module

    Args:

    name (str): name for kernels. example: "2+3mc"
    multihyps (bool, optional): True for using multiple hyperparameter groups

    :return: kernel function, kernel gradient, energy kernel,
             energy_and_force kernel

    """

    #  kernel name should be replace with kernel array
    if component == 'sc':
        stk = sc._str_to_kernel
    else:
        multihyps = True
        if hyps_mask is None:
            multihyps = False
        elif hyps_mask['nspecie'] == 1:
            multihyps = False
        if multihyps:
            stk = mc_sephyps._str_to_kernel
        else:
            stk = mc_simple._str_to_kernel

    # b2 = Two body in use, b3 = Three body in use
    str_terms = {'2': ['2', 'two', 'twobody'],
                 '3': ['3', 'three', 'threebody'],
                 'many': ['mb', 'manybody', 'many']}

    if isinstance(kernels, str):
        kernels = [kernels]

    prefix = ''
    for term in str_terms:
        add = False
        for s in str_terms[term]:
            for k in kernels:
                if s in k.lower():
                    add = True
        if add:
            if len(prefix) > 0:
                prefix += '+'
            prefix += term

    if len(prefix) == 0:
        raise RuntimeError(
            f"the name has to include at least one number {kernels}")

    for suffix in ['', '_grad', '_en', '_force_en', '_efs_energy',
                   '_efs_force', '_efs_self']:
        if prefix+suffix not in stk:
            raise RuntimeError(
                f"cannot find kernel function of {prefix}{suffix}")

    return stk[prefix], stk[prefix + '_grad'], stk[prefix + '_en'], \
        stk[prefix + '_force_en'], stk[prefix+'_efs_energy'], \
        stk[prefix + '_efs_force'], stk[prefix + '_efs_self']



def from_mask_to_args(hyps, cutoffs, hyps_mask=None):
    """ Return the tuple of arguments needed for kernel function.
    The order of the tuple has to be exactly the same as the one taken by
        the kernel function.

    :param hyps: list of hyperparmeter values
    :type hyps: nd.array
    :param hyps_mask: all the remaining parameters needed
    :type hyps_mask: dictionary
    :param cutoffs: cutoffs used

    :return: args
    """

    # no special setting
    multihyps = True
    if hyps_mask is None:
        multihyps = False
    elif hyps_mask['nspecie'] == 1:
        multihyps = False

    if not multihyps:

        cutoffs_array = [0, 0, 0]
        cutoffs_array[0] = cutoffs.get('twobody', 0)
        cutoffs_array[1] = cutoffs.get('threebody', 0)
        cutoffs_array[2] = cutoffs.get('manybody', 0)
        return (hyps, cutoffs_array)

    # setting for mc_sephyps
    nspecie = hyps_mask['nspecie']
    n2b = hyps_mask.get('ntwobody', 0)

    n3b = hyps_mask.get('nthreebody', 0)
    nmanybody = hyps_mask.get('nmanybody', 0)
    ncut3b = hyps_mask.get('ncut3b', 0)

    twobody_mask = hyps_mask.get('twobody_mask', None)
    threebody_mask = hyps_mask.get('threebody_mask', None)
    manybody_mask = hyps_mask.get('manybody_mask', None)
    cut3b_mask = hyps_mask.get('cut3b_mask', None)

    # TO DO , should instead use the non-sephyps kernel
    if (n2b == 1):
        twobody_mask = np.zeros(nspecie**2, dtype=int)
    if (n3b == 1):
        threebody_mask = np.zeros(nspecie**3, dtype=int)
    if (nmanybody == 1):
        manybody_mask = np.zeros(nspecie**2, dtype=int)

    cutoff_2b = cutoffs.get('twobody', 0)
    cutoff_3b = cutoffs.get('threebody', 0)
    cutoff_mb = cutoffs.get('manybody', 0)

    if 'bond_cutoff_list' in hyps_mask:
        cutoff_2b = hyps_mask['bond_cutoff_list']
    else:
        cutoff_2b = np.ones(nspecie**2, dtype=float)*cutoff_2b

    if 'threebody_cutoff_list' in hyps_mask:
        cutoff_3b = hyps_mask['threebody_cutoff_list']
    if 'manybody_cutoff_list' in hyps_mask:
        cutoff_mb = hyps_mask['manybody_cutoff_list']

    (sig2, ls2) = Parameters.get_component_hyps(hyps_mask, 'twobody', hyps=hyps)
    (sig3, ls3) = Parameters.get_component_hyps(
        hyps_mask, 'threebody', hyps=hyps)
    (sigm, lsm) = Parameters.get_component_hyps(
        hyps_mask, 'manybody', hyps=hyps)

    return (cutoff_2b, cutoff_3b, cutoff_mb,
            nspecie,
            np.array(hyps_mask['specie_mask']),
            n2b, twobody_mask,
            n3b, threebody_mask,
            ncut3b, cut3b_mask,
            nmanybody, manybody_mask,
            sig2, ls2, sig3, ls3, sigm, lsm)


def from_grad_to_mask(grad, hyps_mask=None):
    """
    Return gradient which only includes hyperparameters
    which are meant to vary

    :param grad: original gradient vector
    :param hyps_mask: dictionary for hyper-parameters

    :return: newgrad
    """

    constrain = True
    if hyps_mask is None:
        constrain = False
    elif 'map' not in hyps_mask:
        constrain = False
    if not constrain:
        return grad

    hyp_index = hyps_mask['map']

    # setting for mc_sephyps
    # if the last element is not sigma_noise
    if hyp_index[-1] == len(grad):
        hm = hyp_index[:-1]
    else:
        hm = hyp_index

    newgrad = np.zeros(len(hm), dtype=np.float64)
    for i, mapid in enumerate(hm):
        newgrad[i] = grad[mapid]

    return newgrad


def kernel_str_to_array(kernel_name: str):
    """
    Args:

    name (str): name for kernels. example: "2+3mc"

    :return: kernel function, kernel gradient, energy kernel,
             energy_and_force kernel
    """

    #  kernel name should be replace with kernel array
    str_terms = {'twobody': ['2', 'two', 'twobody'],
                 'threebody': ['3', 'three', 'threebody'],
                 'manybody': ['mb', 'manybody', 'many']}

    array = []
    for term in str_terms:
        add = False
        for s in str_terms[term]:
            if s in kernel_name.lower():
                add = True
        if add:
            array += [term]
    return array



"""
Implementation of three-body kernels using different cutoffs.

The kernels are slightly slower.
"""

import numpy as np
import os
import sys

from numba import njit
from math import exp

import flare.kernels.cutoffs as cf

from flare.env import AtomicEnvironment
from flare.kernels.kernels import coordination_number, q_value, q_value_mc, \
    mb_grad_helper_ls_, mb_grad_helper_ls_, k_sq_exp_double_dev, k_sq_exp_dev
from typing import Callable


@njit
def many_body_mc_sepcut_jit(q_array_1, q_array_2,
                            q_neigh_array_1, q_neigh_array_2,
                            q_neigh_grads_1, q_neigh_grads_2,
                            c1, c2, etypes1, etypes2,
                            species1, species2,
                            d1, d2, sig, ls,
                            nspec, spec_mask, mb_mask):
    """many-body multi-element kernel between two force components accelerated
    with Numba.

    Args:
        c1 (int): atomic species of the central atom in env 1
        c2 (int): atomic species of the central atom in env 2
        etypes1 (np.ndarray): atomic species of atoms in env 1
        etypes2 (np.ndarray): atomic species of atoms in env 2
        species1 (np.ndarray): all the atomic species present in trajectory 1
        species2 (np.ndarray): all the atomic species present in trajectory 2
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        sig (float): many-body signal variance hyperparameter.
        ls (float): many-body length scale hyperparameter.

    Return:
        float: Value of the many-body kernel.
    """

    kern = 0

    useful_species = np.array(
        list(set(species1).intersection(set(species2))), dtype=np.int8)

    bc1 = spec_mask[c1]
    bc1n = bc1 * nspec
    bc2 = spec_mask[c2]
    bc2n = bc2 * nspec

    # loop over all possible species
    for s in useful_species:

        bs = spec_mask[s]
        bsn = bs * nspec
        mbtype1 = mb_mask[bc1n + bs]
        mbtype2 = mb_mask[bc2n + bs]

        # Calculate many-body descriptor values for central atoms 1 and 2
        s1 = np.where(species1==s)[0][0]
        s2 = np.where(species2==s)[0][0]
        q1 = q_array_1[s1]
        q2 = q_array_2[s2]

        # compute kernel only if central atoms are of the same species
        if c1 == c2:
            k12 = k_sq_exp_double_dev(q1, q2, sig[mbtype1], ls[mbtype1])
        else:
            k12 = 0

        # initialise arrays of many body descriptors and gradients for the neighbour atoms in
        # the two configurations
        # Loop over neighbours i of 1st configuration
        for i in range(q_neigh_array_1.shape[0]):
            qis = q1i_grads = qi1_grads = ki2s = 0
            if etypes1[i] == s:
                q1i_grads = q_neigh_grads_1[i, d1-1]

            if c1 == s:
                qi1_grads = q_neigh_grads_1[i, d1-1]

            # Calculate many-body descriptor value for i
            qis = q_neigh_array_1[i, s1]

            if c2 == etypes1[i]:
                ki2s = k_sq_exp_double_dev(qis, q2, sig[mbtype2], ls[mbtype2])

            # Loop over neighbours j of 2
            for j in range(q_neigh_array_2.shape[0]):
                qjs = qj2_grads = q2j_grads = k1js = 0

                if etypes2[j] == s:
                    q2j_grads = q_neigh_grads_2[j, d2-1]

                if c2 == s:
                    qj2_grads = q_neigh_grads_2[j, d2-1]

                # Calculate many-body descriptor value for j
                qjs = q_neigh_array_2[j, s2]

                if c1 == etypes2[j]:
                    k1js = k_sq_exp_double_dev(q1, qjs, sig[mbtype1], ls[mbtype1])

                be = spec_mask[etypes1[i]]
                mbtype = mb_mask[be+bsn]
                if etypes1[i] == etypes2[j]:
                    kij = k_sq_exp_double_dev(qis, qjs, sig[mbtype], ls[mbtype])
                else:
                    kij = 0

                kern += q1i_grads * q2j_grads * k12
                kern += qi1_grads * q2j_grads * ki2s
                kern += q1i_grads * qj2_grads * k1js
                kern += qi1_grads * qj2_grads * kij
    return kern

@njit
def many_body_mc_grad_sepcut_jit(q_array_1, q_array_2,
                            q_neigh_array_1, q_neigh_array_2,
                            q_neigh_grads_1, q_neigh_grads_2,
                            c1, c2, etypes1, etypes2,
                            species1, species2,
                            d1, d2, sig, ls,
                            nspec, spec_mask, nmb, mb_mask):
    """gradient of many-body multi-element kernel between two force components
    w.r.t. the hyperparameters, accelerated with Numba.

    Args:
        c1 (int): atomic species of the central atom in env 1
        c2 (int): atomic species of the central atom in env 2
        etypes1 (np.ndarray): atomic species of atoms in env 1
        etypes2 (np.ndarray): atomic species of atoms in env 2
        species1 (np.ndarray): all the atomic species present in trajectory 1
        species2 (np.ndarray): all the atomic species present in trajectory 2
        d1 (int): Force component of the first environment.
        d2 (int): Force component of the second environment.
        sig (float): many-body signal variance hyperparameter.
        ls (float): many-body length scale hyperparameter.

    Return:
        array: Value of the many-body kernel and its gradient w.r.t. sig and ls
    """

    kern = 0
    sig_derv = np.zeros(nmb, dtype=np.float64)
    ls_derv = np.zeros(nmb, dtype=np.float64)

    useful_species = np.array(
        list(set(species1).intersection(set(species2))), dtype=np.int8)

    bc1 = spec_mask[c1]
    bc1n = bc1 * nspec
    bc2 = spec_mask[c2]
    bc2n = bc2 * nspec

    for s in useful_species:

        bs = spec_mask[s]
        bsn = bs * nspec
        mbtype1 = mb_mask[bc1n + bs]
        mbtype2 = mb_mask[bc2n + bs]

        # Calculate many-body descriptor values for central atoms 1 and 2
        s1 = np.where(species1==s)[0][0]
        s2 = np.where(species2==s)[0][0]
        q1 = q_array_1[s1]
        q2 = q_array_2[s2]

        # compute kernel only if central atoms are of the same species
        if c1 == c2:
            k12 = k_sq_exp_double_dev(q1, q2, sig[mbtype1], ls[mbtype1])
            q12diffsq = (q1 - q2) ** 2  # * (q1 - q2)
            dk12 = mb_grad_helper_ls_(q12diffsq, sig[mbtype1], ls[mbtype1])
        else:
            k12 = 0
            dk12 = 0

        # Compute  ki2s, qi1_grads, and qis
        for i in range(q_neigh_array_1.shape[0]):
            qis = q1i_grads = qi1_grads = ki2s = dki2s = 0
            if etypes1[i] == s:
                q1i_grads = q_neigh_grads_1[i, d1-1]

            if c1 == s:
                qi1_grads = q_neigh_grads_1[i, d1-1]

            # Calculate many-body descriptor value for i
            qis = q_neigh_array_1[i, s1]

            if c2 == etypes1[i]:
                ki2s = k_sq_exp_double_dev(qis, q2, sig[mbtype2], ls[mbtype2])
                qi2diffsq = (qis - q2) * (qis - q2)
                dki2s = mb_grad_helper_ls_(qi2diffsq, sig[mbtype2], ls[mbtype2])

            # Compute k1js, qj2_grads and qjs
            for j in range(q_neigh_array_2.shape[0]):
                qjs = qj2_grads = q2j_grads = k1js = dk1js = 0
                if etypes2[j] == s:
                    q2j_grads = q_neigh_grads_2[j, d2-1]

                if c2 == s:
                    qj2_grads = q_neigh_grads_2[j, d2-1]

                # Calculate many-body descriptor value for j
                qjs = q_neigh_array_2[j, s2]

                if c1 == etypes2[j]:
                    k1js = k_sq_exp_double_dev(q1, qjs, sig[mbtype1], ls[mbtype1])
                    q1jdiffsq = (q1 - qjs) * (q1 - qjs)
                    dk1js = mb_grad_helper_ls_(q1jdiffsq, sig[mbtype1], ls[mbtype1])

                be = spec_mask[etypes2[j]]
                mbtype = mb_mask[bsn + be]
                if etypes1[i] == etypes2[j]:
                    kij = k_sq_exp_double_dev(
                        qis, qjs, sig[mbtype], ls[mbtype])
                    qijdiffsq = (qis - qjs) * (qis - qjs)
                    dkij = mb_grad_helper_ls_(
                        qijdiffsq, sig[mbtype], ls[mbtype])
                else:
                    kij = 0
                    dkij = 0

                # c1 s and c2 s and if c1==c2 --> c1 s
                if k12 != 0:
                    kern_term_c1s = q1i_grads * q2j_grads * k12
                    if sig[mbtype1] !=0:
                        sig_derv[mbtype1] += kern_term_c1s * 2. / sig[mbtype1]
                    kern += kern_term_c1s
                    ls_derv[mbtype1] += q1i_grads * q2j_grads * dk12

                # s e1 and c2 s and c2==e1 --> c2 s
                if ki2s != 0:
                    kern_term_c2s = qi1_grads * q2j_grads * ki2s
                    if sig[mbtype2] !=0:
                        sig_derv[mbtype2] += kern_term_c2s * 2. / sig[mbtype2]
                    kern += kern_term_c2s
                    ls_derv[mbtype2] += qi1_grads * q2j_grads * dki2s

                # c1 s and s e2 and  c1==e2 --> c1 s
                if k1js != 0:
                    kern_term_c1s = q1i_grads * qj2_grads * k1js
                    if sig[mbtype1] !=0:
                        sig_derv[mbtype1] += kern_term_c1s * 2. / sig[mbtype1]
                    kern += kern_term_c1s
                    ls_derv[mbtype1] += q1i_grads * qj2_grads * dk1js

                # s e1 and s e2 and e1 == e2 -> s e
                if kij != 0:
                    kern_term_se = qi1_grads * qj2_grads * kij
                    if sig[mbtype] !=0:
                        sig_derv[mbtype] += kern_term_se * 2. / sig[mbtype]
                    kern += kern_term_se
                    ls_derv[mbtype]  += qi1_grads * qj2_grads * dkij


    grad = np.zeros(nmb*2, dtype=np.float64)
    grad[:nmb] = sig_derv
    grad[nmb:] = ls_derv

    return kern, grad


@njit
def many_body_mc_force_en_sepcut_jit(q_array_1, q_array_2,
                                     q_neigh_array_1, q_neigh_grads_1,
                                     c1, c2, etypes1,
                                     species1, species2, d1, sig, ls,
                                     nspec, spec_mask, mb_mask):
    """many-body many-element kernel between force and energy components accelerated
    with Numba.

    Args:
        To be complete
        c1 (int): atomic species of the central atom in env 1
        c2 (int): atomic species of the central atom in env 2
        etypes1 (np.ndarray): atomic species of atoms in env 1
        species1 (np.ndarray): all the atomic species present in trajectory 1
        species2 (np.ndarray): all the atomic species present in trajectory 2
        d1 (int): Force component of the first environment.
        sig (float): many-body signal variance hyperparameter.
        ls (float): many-body length scale hyperparameter.

    Return:
        float: Value of the many-body kernel.
    """

    kern = 0

    useful_species = np.array(
        list(set(species1).intersection(set(species2))), dtype=np.int8)

    bc1 = spec_mask[c1]
    bc1n = bc1 * nspec
    bc2 = spec_mask[c2]
    bc2n = bc2 * nspec

    for s in useful_species:

        bs = spec_mask[s]
        bsn = bs * nspec
        mbtype1 = mb_mask[bc1n + bs]
        mbtype2 = mb_mask[bc2n + bs]

        s1 = np.where(species1==s)[0][0]
        s2 = np.where(species2==s)[0][0]
        q1 = q_array_1[s1]
        q2 = q_array_2[s2]

        if c1 == c2:
            k12 = k_sq_exp_dev(q1, q2, sig[mbtype1], ls[mbtype1])
        else:
            k12 = 0

        # Loop over neighbours i of 1
        for i in range(q_neigh_array_1.shape[0]):
            qi1_grads = q1i_grads = 0
            ki2s = 0

            if etypes1[i] == s:
                q1i_grads = q_neigh_grads_1[i, d1-1]

            if (c1 == s) and (c2 == etypes1[i]):
                qi1_grads = q_neigh_grads_1[i, d1-1]
                qis = q_neigh_array_1[i, s1]
                ki2s = k_sq_exp_dev(qis, q2, sig[mbtype2], ls[mbtype2])

            kern -= q1i_grads * k12 + qi1_grads * ki2s
    return kern


@njit
def many_body_mc_en_sepcut_jit(q_array_1, q_array_2, c1, c2,
                               species1, species2,
                               sig, ls,
                               nspec, spec_mask, mb_mask):
    """many-body many-element kernel between energy components accelerated
    with Numba.

    Args:
        bond_array_1 (np.ndarray): many-body bond array of the first local
            environment.
        bond_array_2 (np.ndarray): many-body bond array of the second local
            environment.
        c1 (int): atomic species of the central atom in env 1
        c2 (int): atomic species of the central atom in env 2
        etypes1 (np.ndarray): atomic species of atoms in env 1
        etypes2 (np.ndarray): atomic species of atoms in env 2
        species1 (np.ndarray): all the atomic species present in trajectory 1
        species2 (np.ndarray): all the atomic species present in trajectory 2
        sig (float): many-body signal variance hyperparameter.
        ls (float): many-body length scale hyperparameter.
        r_cut (float): many-body cutoff radius.
        cutoff_func (Callable): Cutoff function.

    Return:
        float: Value of the many-body kernel.
    """
    useful_species = np.array(
        list(set(species1).intersection(set(species2))), dtype=np.int8)

    kern = 0

    if c1 == c2:
        ls2 = ls*ls
        sig2 = sig*sig

        bc1 = spec_mask[c1]
        bc1n = bc1 * nspec

        for s in useful_species:

            bs = spec_mask[s]
            mbtype = mb_mask[bc1n + bs]

            tls2 = ls2[mbtype]
            tsig2 = sig2[mbtype]

            q1 = q_array_1[np.where(species1==s)[0][0]]
            q2 = q_array_2[np.where(species2==s)[0][0]]

            q1q2diff = q1 - q2

            kern += tsig2 * exp(-q1q2diff * q1q2diff / (2 * tls2))

    return kern



from flare.struc import Structure
from typing import List
from json import dump, load
from flare.utils.element_coder import NumpyEncoder

def md_trajectory_to_file(filename: str, structures: List[Structure]):
    """
    Take a list of structures and write them to a json file.
    :param filename:
    :param structures:
    """
    with open(filename, 'w') as f:
        dump([s.as_dict() for s in structures], f, cls=NumpyEncoder)

def md_trajectory_from_file(filename: str):
    """
    Read a list of structures from a json file, formatted as in md_trajectory_to_file.
    :param filename:
    """
    with open(filename, 'r') as f:
        structure_list = load(f)
        structures = \
            [Structure.from_dict(dictionary) for dictionary in structure_list]
    return structures
